# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import re

from gi.repository import Adw, Gdk, GObject, Gtk

from dialect.define import RES_PATH
from dialect.languages import LangObject, LanguagesListModel


@Gtk.Template(resource_path=f"{RES_PATH}/widgets/lang_selector.ui")
class LangSelector(Adw.Bin):
    __gtype_name__ = "LangSelector"

    # Properties
    selected: str = GObject.Property(type=str)  # type: ignore

    # Child Widgets
    button: Gtk.MenuButton = Gtk.Template.Child()  # type: ignore
    label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    insight: Gtk.Label = Gtk.Template.Child()  # type: ignore
    popover: Gtk.Popover = Gtk.Template.Child()  # type: ignore
    search: Gtk.SearchEntry = Gtk.Template.Child()  # type: ignore
    scroll: Gtk.ScrolledWindow = Gtk.Template.Child()  # type: ignore
    revealer: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    recent_list: Gtk.ListBox = Gtk.Template.Child()  # type: ignore
    separator: Gtk.Separator = Gtk.Template.Child()  # type: ignore
    lang_list: Gtk.ListBox = Gtk.Template.Child()  # type: ignore

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.model: LanguagesListModel | None = None
        self.recent_model: LanguagesListModel | None = None

        # Setup search entry
        self.search.set_key_capture_widget(self.popover)
        key_events = Gtk.EventControllerKey()
        key_events.connect("key-pressed", self._on_key_pressed)
        self.search.add_controller(key_events)

    @GObject.Signal()
    def user_selection_changed(self): ...

    def bind_models(self, langs: LanguagesListModel, recent: LanguagesListModel):
        self.model = langs
        self.recent_model = recent

        self.recent_model.connect("items-changed", self._on_recent_changed)

        self.filter = Gtk.CustomFilter()
        self.filter.set_filter_func(self._filter_langs)
        sorter = Gtk.CustomSorter()
        sorter.set_sort_func(self._sort_langs)
        sorted_model = Gtk.SortListModel(model=self.model, sorter=sorter)
        filter_model = Gtk.FilterListModel(model=sorted_model, filter=self.filter)
        self.lang_list.bind_model(filter_model, self._create_lang_row)

        self.recent_list.bind_model(self.recent_model, self._create_lang_row)

    def set_insight(self, code: str):
        if self.selected == "auto":
            self.insight.props.label = f"({self._get_lang_name(code)})"

    def _get_lang_name(self, code: str) -> str:
        if self.model:
            return self.model.names_func(code) or code
        return code

    def _on_recent_changed(self, _model, _position: int, _removed: int, _added: int):
        if self.recent_model:
            self.recent_model.set_selected(self.selected)

    @Gtk.Template.Callback()
    def _on_selected_changed(self, _self, _pspec):
        """Called on self::notify::selected signal"""

        if self.model is not None:
            self.model.set_selected(self.selected)

            if self.selected == "auto":
                self.label.props.label = _("Auto")
            else:
                self.label.props.label = self._get_lang_name(self.selected)

            self.insight.props.label = ""

    @Gtk.Template.Callback()
    def _activated(self, _list, row: "LangRow"):
        """Called on self.(recent_list, lang_list)::row-activated signal"""
        # Close popover
        self.popover.popdown()
        # Set selected property
        self.selected = row.lang.code
        self.emit("user-selection-changed")

    @Gtk.Template.Callback()
    def _popover_show(self, _popover):
        """Called on self.popover::show signal"""
        self.search.grab_focus()

    @Gtk.Template.Callback()
    def _popover_closed(self, _popover):
        """Called on self.popover::closed signal"""
        # Reset scroll
        vscroll = self.scroll.get_vadjustment()
        vscroll.props.value = 0
        # Clear search
        self.search.props.text = ""

    def _create_lang_row(self, lang: LangObject):
        return LangRow(lang)

    def _filter_langs(self, item):
        search = self.search.get_text()
        return bool(re.search(search, item.name, re.IGNORECASE))

    def _sort_langs(self, lang_a, lang_b, _data):
        a = lang_a.name.lower()
        b = lang_b.name.lower()
        return (a > b) - (a < b)

    @Gtk.Template.Callback()
    def _on_search(self, _entry):
        """Called on self.search::changed signal"""
        if self.search.props.text != "":
            self.revealer.props.reveal_child = False
        else:
            self.revealer.props.reveal_child = True

        self.filter.emit("changed", Gtk.FilterChange.DIFFERENT)

    @Gtk.Template.Callback()
    def _on_search_activate(self, _entry):
        """Called on self.search::activate signal"""
        if self.search.props.text:
            row = self.lang_list.get_row_at_index(0)
            if row:
                self.lang_list.emit("row-activated", row)
        return Gdk.EVENT_PROPAGATE

    def _on_key_pressed(self, _ctrl, keyval: int, _keycode: int, _mod: Gdk.ModifierType):
        # Close popover if ESQ key is pressed in search entry
        if keyval == Gdk.KEY_Escape:
            self.popover.popdown()
        # Prevent search entry unfocusing when down key is pressed
        elif keyval == Gdk.KEY_Down:
            return Gdk.EVENT_STOP


@Gtk.Template(resource_path=f"{RES_PATH}/widgets/lang_row.ui")
class LangRow(Gtk.ListBoxRow):
    __gtype_name__ = "LangRow"

    # Widgets
    name: Gtk.Label = Gtk.Template.Child()  # type: ignore
    selection: Gtk.Image = Gtk.Template.Child()  # type: ignore

    def __init__(self, lang: LangObject):
        super().__init__()
        self.lang = lang
        self.name.props.label = self.lang.name

        self.lang.bind_property("selected", self.selection, "visible", GObject.BindingFlags.SYNC_CREATE)



# Copyright 2023 Mufeed Ali
# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import threading
from typing import Callable

from dialect.providers.base import BaseProvider


class LocalProvider(BaseProvider):
    """Base class for providers needing local threaded helpers"""

    def launch_thread(self, worker: Callable, *args):
        """
        Launches a thread using Python's threading.

        Args:
            worker: Function to execute on the thread
            *args: Args for the worker
        """
        threading.Thread(target=worker, args=args, daemon=True).start()



# Copyright 2020 gi-lom
# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# Copyright 2023 Libretto
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from typing import IO, Literal, TypedDict

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gst, Gtk

from dialect.define import APP_ID, PROFILE, RES_PATH, TRANS_NUMBER
from dialect.languages import LanguagesListModel
from dialect.providers import (
    TRANSLATORS,
    TTS,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
)
from dialect.providers.base import BaseProvider, Translation
from dialect.settings import Settings
from dialect.shortcuts import DialectShortcutsWindow
from dialect.utils import find_item_match, first_exclude
from dialect.widgets import LangSelector, SpeechButton, TextView, ThemeSwitcher


class _OngoingSpeech(TypedDict):
    text: str
    lang: str
    called_from: Literal["src", "dest"]


class _NotificationAction(TypedDict):
    label: str
    name: str


@Gtk.Template(resource_path=f"{RES_PATH}/window.ui")
class DialectWindow(Adw.ApplicationWindow):
    __gtype_name__ = "DialectWindow"

    # Properties
    translator_loading: bool = GObject.Property(type=bool, default=True)  # type: ignore

    # Child widgets
    menu_btn: Gtk.MenuButton = Gtk.Template.Child()  # type: ignore
    main_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    error_page: Adw.StatusPage = Gtk.Template.Child()  # type: ignore
    translator_box: Gtk.Box = Gtk.Template.Child()  # type: ignore
    key_page: Adw.StatusPage = Gtk.Template.Child()  # type: ignore
    rmv_key_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    error_api_key_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore

    title_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    langs_button_box: Gtk.Box = Gtk.Template.Child()  # type: ignore
    switch_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    src_lang_selector: LangSelector = Gtk.Template.Child()  # type: ignore
    dest_lang_selector: LangSelector = Gtk.Template.Child()  # type: ignore

    return_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    forward_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore

    src_pron_revealer: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    src_pron_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    mistakes: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    mistakes_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    char_counter: Gtk.Label = Gtk.Template.Child()  # type: ignore
    src_text: TextView = Gtk.Template.Child()  # type: ignore
    clear_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    paste_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    src_speech_btn: SpeechButton = Gtk.Template.Child()  # type: ignore
    translate_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore

    dest_box: Gtk.Box = Gtk.Template.Child()  # type: ignore
    dest_pron_revealer: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    dest_pron_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    dest_text: TextView = Gtk.Template.Child()  # type: ignore
    dest_toolbar_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    trans_spinner: Adw.Spinner = Gtk.Template.Child()  # type: ignore
    trans_warning: Gtk.Image = Gtk.Template.Child()  # type: ignore
    edit_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    copy_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    dest_speech_btn: SpeechButton = Gtk.Template.Child()  # type: ignore

    actionbar: Gtk.ActionBar = Gtk.Template.Child()  # type: ignore
    src_lang_selector_m: LangSelector = Gtk.Template.Child()  # type: ignore
    dest_lang_selector_m: LangSelector = Gtk.Template.Child()  # type: ignore

    toast: Adw.Toast | None = None  # for notification management
    toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()  # type: ignore

    win_key_ctrlr: Gtk.EventControllerKey = Gtk.Template.Child()  # type: ignore

    # Providers objects
    provider: dict[str, BaseProvider | None] = {"trans": None, "tts": None}

    # Text to speech
    speech_provider_failed = False  # tts provider loading failed
    current_speech: _OngoingSpeech | None = None
    speech_loading = False  # tts loading status

    # Preset language values
    src_langs: list[str] = []
    dest_langs: list[str] = []

    current_history = 0  # for history management

    # Translation-related variables
    next_trans = {}  # for ongoing translation
    ongoing_trans = False  # for ongoing translation
    trans_mistakes: tuple[str | None, str | None] = (None, None)  # "mistakes" suggestions
    # Pronunciations
    trans_src_pron = None
    trans_dest_pron = None
    # Suggestions
    before_suggest = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Application object
        self.app: Adw.Application = kwargs["application"]

        # GStreamer playbin object and related setup
        self.player = Gst.ElementFactory.make("playbin", "player")
        if self.player:
            if bus := self.player.get_bus():
                bus.add_signal_watch()
                bus.connect("message", self._on_gst_message)

        # Setup window
        self.setup_actions()
        self.setup()

    def setup_actions(self):
        back = Gio.SimpleAction(name="back")
        back.props.enabled = False
        back.connect("activate", self.ui_return)
        self.add_action(back)

        forward_action = Gio.SimpleAction(name="forward")
        forward_action.props.enabled = False
        forward_action.connect("activate", self.ui_forward)
        self.add_action(forward_action)

        switch_action = Gio.SimpleAction(name="switch")
        switch_action.connect("activate", self.ui_switch)
        self.add_action(switch_action)

        from_action = Gio.SimpleAction(name="from")
        from_action.connect("activate", self.ui_from)
        self.add_action(from_action)

        to_action = Gio.SimpleAction(name="to")
        to_action.connect("activate", self.ui_to)
        self.add_action(to_action)

        clear_action = Gio.SimpleAction(name="clear")
        clear_action.props.enabled = False
        clear_action.connect("activate", self.ui_clear)
        self.add_action(clear_action)

        font_size_inc_action = Gio.SimpleAction(name="font-size-inc")
        font_size_inc_action.connect("activate", self.ui_font_size_inc)
        self.add_action(font_size_inc_action)

        font_size_dec_action = Gio.SimpleAction(name="font-size-dec")
        font_size_dec_action.connect("activate", self.ui_font_size_dec)
        self.add_action(font_size_dec_action)

        paste_action = Gio.SimpleAction(name="paste")
        paste_action.connect("activate", self.ui_paste)
        self.add_action(paste_action)

        copy_action = Gio.SimpleAction(name="copy")
        copy_action.props.enabled = False
        copy_action.connect("activate", self.ui_copy)
        self.add_action(copy_action)

        listen_dest_action = Gio.SimpleAction(name="listen-dest")
        listen_dest_action.connect("activate", self.ui_dest_listen)
        listen_dest_action.props.enabled = False
        self.add_action(listen_dest_action)

        suggest_action = Gio.SimpleAction(name="suggest")
        suggest_action.props.enabled = False
        suggest_action.connect("activate", self.ui_suggest)
        self.add_action(suggest_action)

        suggest_ok_action = Gio.SimpleAction(name="suggest-ok")
        suggest_ok_action.connect("activate", self.ui_suggest_ok)
        self.add_action(suggest_ok_action)

        suggest_cancel_action = Gio.SimpleAction(name="suggest-cancel")
        suggest_cancel_action.connect("activate", self.ui_suggest_cancel)
        self.add_action(suggest_cancel_action)

        listen_src_action = Gio.SimpleAction(name="listen-src")
        listen_src_action.connect("activate", self.ui_src_listen)
        listen_src_action.props.enabled = False
        self.add_action(listen_src_action)

        translation_action = Gio.SimpleAction(name="translation")
        translation_action.props.enabled = False
        translation_action.connect("activate", self.translation)
        self.add_action(translation_action)

    def setup(self):
        self.set_default_icon_name(APP_ID)

        # Set devel style
        if PROFILE == "Devel":
            self.add_css_class("devel")

        # Theme Switcher
        theme_switcher = ThemeSwitcher()
        menu: Gtk.PopoverMenu = self.menu_btn.props.popover  # type: ignore
        menu.add_child(theme_switcher, "theme")

        # Save settings on close
        self.connect("unrealize", self.save_settings)

        self.setup_selectors()
        self.setup_translation()
        self.set_help_overlay(DialectShortcutsWindow())

        # Load translator
        self.load_translator()
        # Load text to speech
        self.load_tts()

        # Listen to active providers changes
        Settings.get().connect("translator-changed", self._on_active_provider_changed, "trans")
        Settings.get().connect("tts-changed", self._on_active_provider_changed, "tts")

        # Bind text views font size
        self.src_text.bind_property("font-size", self.dest_text, "font-size", GObject.BindingFlags.BIDIRECTIONAL)

        # Set initial saved text view font size
        if Settings.get().custom_default_font_size:
            font_size = Settings.get().default_font_size
            self.set_font_size(font_size)

        # Set src textview mod key requirement
        self.src_text.activate_mod = not bool(Settings.get().translate_accel_value)
        Settings.get().connect(
            "changed::translate-accel",
            lambda s, _k: self.src_text.set_property("activate_mod", not bool(s.translate_accel_value)),
        )

    def setup_selectors(self):
        def lang_names_func(code: str):
            return self.provider["trans"].get_lang_name(code)  # type: ignore

        # Languages models
        self.src_lang_model = LanguagesListModel(lang_names_func)
        self.src_recent_lang_model = LanguagesListModel(lang_names_func)
        self.dest_lang_model = LanguagesListModel(lang_names_func)
        self.dest_recent_lang_model = LanguagesListModel(lang_names_func)

        # Src lang selector
        self.src_lang_selector.bind_models(self.src_lang_model, self.src_recent_lang_model)
        self.src_lang_selector_m.bind_models(self.src_lang_model, self.src_recent_lang_model)

        # Dest lang selector
        self.dest_lang_selector.bind_models(self.dest_lang_model, self.dest_recent_lang_model)
        self.dest_lang_selector_m.bind_models(self.dest_lang_model, self.dest_recent_lang_model)

        self.langs_button_box.props.homogeneous = False

    def setup_translation(self):
        # Src buffer
        self.src_buffer = self.src_text.props.buffer
        self.src_buffer.connect("changed", self.on_src_text_changed)
        self.src_buffer.connect("end-user-action", self.user_action_ended)

        # Dest buffer
        self.dest_buffer = self.dest_text.props.buffer
        self.dest_buffer.props.text = ""
        self.dest_buffer.connect("changed", self.on_dest_text_changed)
        # Translation progress spinner
        self.trans_spinner.hide()
        self.trans_warning.hide()

    def load_translator(self):
        def on_done():
            if not self.provider["trans"]:
                return

            # Mistakes support
            if ProviderFeature.MISTAKES not in self.provider["trans"].features:
                self.mistakes.props.reveal_child = False

            # Suggestions support
            self.ui_suggest_cancel(None, None)
            if ProviderFeature.SUGGESTIONS not in self.provider["trans"].features:
                self.edit_btn.props.visible = False
            else:
                self.edit_btn.props.visible = True

            # Pronunciation support
            if ProviderFeature.PRONUNCIATION not in self.provider["trans"].features:
                self.src_pron_revealer.props.reveal_child = False
                self.dest_pron_revealer.props.reveal_child = False
                self.app.lookup_action("pronunciation").props.enabled = False  # type: ignore
            else:
                self.app.lookup_action("pronunciation").props.enabled = True  # type: ignore

            # Update langs
            self.src_lang_model.set_langs(self.provider["trans"].src_languages)
            self.dest_lang_model.set_langs(self.provider["trans"].dest_languages)

            # Update selected langs
            set_auto = Settings.get().src_auto and ProviderFeature.DETECTION in self.provider["trans"].features
            src_lang = self.provider["trans"].src_languages[0]
            if self.src_langs and self.src_langs[0] in self.provider["trans"].src_languages:
                src_lang = self.src_langs[0]
            self.src_lang_selector.selected = "auto" if set_auto else src_lang

            dest_lang = self.provider["trans"].dest_languages[1]
            if self.dest_langs and self.dest_langs[0] in self.provider["trans"].dest_languages:
                dest_lang = self.dest_langs[0]
            self.dest_lang_selector.selected = dest_lang

            # Update chars limit
            if self.provider["trans"].chars_limit == -1:  # -1 means unlimited
                self.char_counter.props.label = ""
            else:
                count = f"{str(self.src_buffer.get_char_count())}/{self.provider['trans'].chars_limit}"
                self.char_counter.props.label = count

            self.translator_loading = False

            self.check_apikey()

        def on_fail(error: ProviderError):
            self.translator_loading = False
            self.loading_failed(error)

        provider = Settings.get().active_translator

        # Show loading view
        self.main_stack.props.visible_child_name = "loading"

        # Translator object
        self.provider["trans"] = TRANSLATORS[provider]()
        # Get saved languages
        self.src_langs = self.provider["trans"].recent_src_langs
        self.dest_langs = self.provider["trans"].recent_dest_langs
        # Do provider init
        self.provider["trans"].init_trans(on_done, on_fail)

        # Connect to provider settings changes
        self.provider["trans"].settings.connect(
            "changed::instance-url", self._on_provider_changed, self.provider["trans"].name
        )
        self.provider["trans"].settings.connect(
            "changed::api-key", self._on_provider_changed, self.provider["trans"].name
        )

    def check_apikey(self):
        def on_done(valid: bool):
            if valid:
                self.main_stack.props.visible_child_name = "translate"
            else:
                self.api_key_failed()

        def on_fail(error: ProviderError):
            self.loading_failed(error)

        if not self.provider["trans"]:
            return

        if ProviderFeature.API_KEY in self.provider["trans"].features:
            if self.provider["trans"].api_key:
                self.provider["trans"].validate_api_key(self.provider["trans"].api_key, on_done, on_fail)
            elif (
                not self.provider["trans"].api_key
                and ProviderFeature.API_KEY_REQUIRED in self.provider["trans"].features
            ):
                self.api_key_failed(required=True)
            else:
                self.main_stack.props.visible_child_name = "translate"
        else:
            self.main_stack.props.visible_child_name = "translate"

    def loading_failed(self, error: ProviderError):
        if not self.provider["trans"]:
            return

        # Api Key error
        if error.code in (ProviderErrorCode.API_KEY_INVALID, ProviderErrorCode.API_KEY_REQUIRED):
            self.api_key_failed(error.code == ProviderErrorCode.API_KEY_REQUIRED)

        # Other errors
        else:
            self.main_stack.props.visible_child_name = "error"

            service = self.provider["trans"].prettyname
            url = self.provider["trans"].instance_url

            title = _("Failed loading the translation service")
            description = _("Please report this in the Dialect bug tracker if the issue persists.")
            if ProviderFeature.INSTANCES in self.provider["trans"].features:
                description = _(
                    (
                        'Failed loading "{url}", check if the instance address is correct or report in the Dialect bug tracker'
                        " if the issue persists."
                    )
                )
                description = description.format(url=url)

            if error.code == ProviderErrorCode.NETWORK:
                title = _("Couldn’t connect to the translation service")
                description = _("We can’t connect to the server. Please check for network issues.")
                if ProviderFeature.INSTANCES in self.provider["trans"].features:
                    description = _(
                        (
                            "We can’t connect to the {service} instance “{url}”.\n"
                            "Please check for network issues or if the address is correct."
                        )
                    )
                    description = description.format(service=service, url=url)

            if error.message:
                description = description + "\n\n<small><tt>" + error.message + "</tt></small>"

            self.error_page.props.title = title
            self.error_page.props.description = description

    def api_key_failed(self, required=False):
        if not self.provider["trans"]:
            return

        if required:
            self.key_page.props.title = _("API key is required to use the service")
            self.key_page.props.description = _("Please set an API key in the preferences.")

        else:
            self.key_page.props.title = _("The provided API key is invalid")
            if ProviderFeature.API_KEY_REQUIRED in self.provider["trans"].features:
                self.key_page.props.description = _("Please set a valid API key in the preferences.")
            else:
                self.key_page.props.description = _(
                    "Please set a valid API key or unset the API key in the preferences."
                )
                self.rmv_key_btn.props.visible = True
                self.error_api_key_btn.props.visible = True

        self.main_stack.props.visible_child_name = "api-key"

    @Gtk.Template.Callback()
    def retry_load_translator(self, _button):
        self.load_translator()

    @Gtk.Template.Callback()
    def remove_key_and_reload(self, _button):
        if self.provider["trans"]:
            self.provider["trans"].reset_api_key()
        self.load_translator()

    def load_tts(self):
        def on_done():
            self.speech_provider_failed = False
            self.src_speech_btn.ready()
            self.dest_speech_btn.ready()
            self._check_speech_enabled()

        def on_fail(error: ProviderError):
            button_text = _("Failed loading the text-to-speech service. Retry?")
            toast_text = _("Failed loading the text-to-speech service")
            if error.code == ProviderErrorCode.NETWORK:
                toast_text = _("Failed loading the text-to-speech service, check for network issues")

            self.speech_provider_failed = True
            self.src_speech_btn.error(button_text)
            self.dest_speech_btn.error(button_text)
            self.send_notification(toast_text)
            self._check_speech_enabled()

        self.src_speech_btn.loading()
        self.dest_speech_btn.loading()

        # TTS name
        provider = Settings.get().active_tts

        # Check if TTS is disabled
        if provider != "":
            self.src_speech_btn.props.visible = True
            self.dest_speech_btn.props.visible = True

            # TTS Object
            self.provider["tts"] = TTS[provider]()
            self.provider["tts"].init_tts(on_done, on_fail)

            # Connect to provider settings changes
            self.provider["tts"].settings.connect(
                "changed::instance-url", self._on_provider_changed, self.provider["tts"].name
            )
            self.provider["tts"].settings.connect(
                "changed::api-key", self._on_provider_changed, self.provider["tts"].name
            )
        else:
            self.provider["tts"] = None
            self.src_speech_btn.props.visible = False
            self.dest_speech_btn.props.visible = False

    def translate(self, text: str, src_lang: str | None, dest_lang: str | None):
        """
        Translates the given text from auto detected language to last used
        language
        """
        if not self.provider["trans"]:
            return

        # Set src lang to Auto
        if src_lang is None:
            self.src_lang_selector.selected = "auto"
        else:
            self.src_lang_selector.selected = src_lang
        if dest_lang is not None and dest_lang in self.provider["trans"].dest_languages:
            self.dest_lang_selector.selected = dest_lang
            self.dest_lang_selector.emit("user-selection-changed")
        # Set text to src buffer
        self.src_buffer.props.text = text
        # Run translation
        self.translation()

    def translate_selection(self, src_lang: str | None, dest_lang: str | None):
        def on_paste(clipboard, result):
            text = clipboard.read_text_finish(result)
            self.translate(text, src_lang, dest_lang)

        if display := Gdk.Display.get_default():
            display.get_primary_clipboard().read_text_async(None, on_paste)

    def save_settings(self, *args, **kwargs):
        if not self.is_maximized():
            size = self.get_default_size()
            Settings.get().window_size = (size.width, size.height)  # type: ignore
        if self.provider["trans"] is not None:
            self.provider["trans"].recent_src_langs = self.src_langs
            self.provider["trans"].recent_dest_langs = self.dest_langs

    def send_notification(
        self,
        text: str,
        queue: bool | None = False,
        action: _NotificationAction | None = None,
        timeout=5,
        priority=Adw.ToastPriority.NORMAL,
    ):
        """
        Display an in-app notification.

        Args:
            text: The text or message of the notification.
            queue: If True, the notification will be queued.
            action: A dict containing the action to be called.
            timeout: Toast timeout.
            timeout: Toast priority.
        """

        def toast_dismissed(_toast: Adw.Toast):
            self.toast = None

        if not queue and self.toast is not None:
            self.toast.dismiss()
        self.toast = Adw.Toast(title=text)
        self.toast.connect("dismissed", toast_dismissed)
        if action is not None:
            self.toast.props.button_label = action["label"]
            self.toast.props.action_name = action["name"]
        self.toast.props.timeout = timeout
        self.toast.props.priority = priority
        self.toast_overlay.add_toast(self.toast)

    def _check_speech_enabled(self):
        if not self.provider["tts"]:
            return

        src_playing = dest_playing = False
        if self.current_speech:
            src_playing = self.current_speech["called_from"] == "src"
            dest_playing = self.current_speech["called_from"] == "dest"

        # Check src listen button
        self.lookup_action("listen-src").set_enabled(  # type: ignore
            self.speech_provider_failed
            or self.src_lang_selector.selected in self.provider["tts"].tts_languages
            and self.src_buffer.get_char_count() != 0
            and not self.speech_loading
            and not dest_playing
        )

        # Check dest listen button
        self.lookup_action("listen-dest").set_enabled(  # type: ignore
            self.speech_provider_failed
            or self.dest_lang_selector.selected in self.provider["tts"].tts_languages
            and self.dest_buffer.get_char_count() != 0
            and not self.speech_loading
            and not src_playing
        )

    @Gtk.Template.Callback()
    def _on_src_lang_changed(self, _obj, _param):
        """Called on self.src_lang_selector::notify::selected signal"""
        if not self.provider["trans"]:
            return

        code = self.src_lang_selector.selected
        dest_code = self.dest_lang_selector.selected

        if self.provider["trans"].cmp_langs(code, dest_code):
            # Get first lang from saved src langs that is not current dest
            if valid := first_exclude(self.src_langs, dest_code):
                # Check if it's a valid dest lang
                valid = find_item_match([valid], self.provider["trans"].dest_languages)
            if not valid:  # If not, just get the first lang from the list that is not selected
                valid = first_exclude(self.provider["trans"].dest_languages, dest_code)

            self.dest_lang_selector.selected = valid or ""

        if code in self.provider["trans"].src_languages:
            # Update saved src langs list
            if code in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code)
            elif code.lower() in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code.lower())
            elif len(self.src_langs) == 4:
                self.src_langs.pop()
            self.src_langs.insert(0, code)

        # Rewrite recent langs
        self.src_recent_lang_model.set_langs(self.src_langs, auto=True)

        self._check_switch_enabled()
        self._check_speech_enabled()

    @Gtk.Template.Callback()
    def _on_dest_lang_changed(self, _obj, _param):
        """Called on self.dest_lang_selector::notify::selected signal"""
        if not self.provider["trans"]:
            return

        code = self.dest_lang_selector.selected
        src_code = self.src_lang_selector.selected

        if self.provider["trans"].cmp_langs(code, src_code):
            # Get first lang from saved dest langs that is not current src
            if valid := first_exclude(self.dest_langs, src_code):
                # Check if it's a valid src lang
                valid = find_item_match([valid], self.provider["trans"].src_languages)
            if not valid:  # If not, just get the first lang from the list that is not selected
                valid = first_exclude(self.provider["trans"].src_languages, src_code)

            self.src_lang_selector.selected = valid or ""

        # Update saved dest langs list
        if code in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code)
        elif code.lower() in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code.lower())
        elif len(self.src_langs) == 4:
            self.dest_langs.pop()
        self.dest_langs.insert(0, code)

        # Rewrite recent langs
        self.dest_recent_lang_model.set_langs(self.dest_langs)

        self._check_switch_enabled()
        self._check_speech_enabled()

    def _check_switch_enabled(self):
        if not self.provider["trans"]:
            return

        # Disable or enable switch function.
        self.lookup_action("switch").props.enabled = (  # type: ignore
            self.src_lang_selector.selected in self.provider["trans"].dest_languages
            and self.dest_lang_selector.selected in self.provider["trans"].src_languages
        )

    """
    User interface functions
    """

    def ui_return(self, _action, _param):
        """Go back one step in history."""
        if self.current_history != TRANS_NUMBER:
            self.current_history += 1
            self.history_update()

    def ui_forward(self, _action, _param):
        """Go forward one step in history."""
        if self.current_history != 0:
            self.current_history -= 1
            self.history_update()

    def add_history_entry(self, translation: Translation):
        """Add a history entry to the history list."""
        if not self.provider["trans"]:
            return

        if self.current_history > 0:
            del self.provider["trans"].history[: self.current_history]
            self.current_history = 0
        if len(self.provider["trans"].history) == TRANS_NUMBER:
            self.provider["trans"].history.pop()
        self.provider["trans"].history.insert(0, translation)
        GLib.idle_add(self.reset_return_forward_btns)

    def switch_all(self, src_language: str, dest_language: str, src_text: str, dest_text: str):
        self.src_lang_selector.selected = dest_language
        self.dest_lang_selector.selected = src_language
        self.src_buffer.props.text = dest_text
        self.dest_buffer.props.text = src_text
        self.add_history_entry(Translation(src_text, (dest_text, src_language, dest_language)))

        # Re-enable widgets
        self.langs_button_box.props.sensitive = True
        self.lookup_action("translation").props.enabled = self.src_buffer.get_char_count() != 0  # type: ignore

    def ui_switch(self, _action, _param):
        # Get variables
        self.langs_button_box.props.sensitive = False
        self.lookup_action("translation").props.enabled = False  # type: ignore
        src_language = self.src_lang_selector.selected
        dest_language = self.dest_lang_selector.selected
        src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)
        if src_language == "auto":
            return

        # Switch all
        self.switch_all(src_language, dest_language, src_text, dest_text)

    def ui_from(self, _action, _param):
        self.src_lang_selector.button.popup()

    def ui_to(self, _action, _param):
        self.dest_lang_selector.button.popup()

    def ui_clear(self, _action, _param):
        self.src_buffer.props.text = ""
        self.src_buffer.emit("end-user-action")

    def set_font_size(self, size: int):
        self.src_text.font_size = size

    def ui_font_size_inc(self, _action, _param):
        self.src_text.font_size_inc()

    def ui_font_size_dec(self, _action, _param):
        self.src_text.font_size_dec()

    def ui_copy(self, _action, _param):
        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)
        if display := Gdk.Display.get_default():
            display.get_clipboard().set(dest_text)
            self.send_notification(_("Copied to clipboard"), timeout=1)

    def ui_paste(self, _action, _param):
        def on_paste(clipboard: Gdk.Clipboard, result: Gio.AsyncResult):
            text = clipboard.read_text_finish(result)
            if text is not None:
                end_iter = self.src_buffer.get_end_iter()
                self.src_buffer.insert(end_iter, text)
                self.src_buffer.emit("end-user-action")

        if display := Gdk.Display.get_default():
            display.get_clipboard().read_text_async(None, on_paste)

    def ui_suggest(self, _action, _param):
        self.dest_toolbar_stack.props.visible_child_name = "edit"
        self.before_suggest = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True
        )
        self.dest_text.props.editable = True

    def ui_suggest_ok(self, _action, _param):
        def on_done(success):
            self.dest_toolbar_stack.props.visible_child_name = "default"

            if success:
                self.send_notification(_("New translation has been suggested!"))
            else:
                self.send_notification(_("Suggestion failed."))

            self.dest_text.props.editable = False

        def on_fail(error: ProviderError):
            self.dest_toolbar_stack.props.visible_child_name = "default"
            self.send_notification(_("Suggestion failed."))
            self.dest_text.props.editable = False

        if not self.provider["trans"]:
            return

        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)

        src, dest = self.provider["trans"].denormalize_lang(
            self.provider["trans"].history[self.current_history].original[1],
            self.provider["trans"].history[self.current_history].original[2],
        )

        self.provider["trans"].suggest(
            self.provider["trans"].history[self.current_history].original[0], src, dest, dest_text, on_done, on_fail
        )

        self.before_suggest = None

    def ui_suggest_cancel(self, _action, _param):
        self.dest_toolbar_stack.props.visible_child_name = "default"
        if self.before_suggest is not None:
            self.dest_buffer.props.text = self.before_suggest
            self.before_suggest = None
        self.dest_text.props.editable = False

    def ui_src_listen(self, _action, _param):
        if self.current_speech:
            self._speech_reset()
            return

        src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
        src_language = self.src_lang_selector.selected
        self._pre_speech(src_text, src_language, "src")

    def ui_dest_listen(self, _action, _param):
        if self.current_speech:
            self._speech_reset()
            return

        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)
        dest_language = self.dest_lang_selector.selected
        self._pre_speech(dest_text, dest_language, "dest")

    def _pre_speech(self, text: str, lang: str, called_from: Literal["src", "dest"]):
        if text != "":
            self.speech_loading = True
            self.current_speech = {"text": text, "lang": lang, "called_from": called_from}
            self._check_speech_enabled()

            if self.speech_provider_failed:
                self.load_tts()
            else:
                self._download_speech()

            if called_from == "src":  # Show spinner on button
                self.src_speech_btn.loading()
            else:
                self.dest_speech_btn.loading()
        elif self.speech_provider_failed:
            self.load_tts()

    def _speech_reset(self, set_ready: bool = True):
        if not self.player:
            return

        self.player.set_state(Gst.State.NULL)
        self.current_speech = None
        self.speech_loading = False
        self._check_speech_enabled()

        if set_ready:
            self.src_speech_btn.ready()
            self.dest_speech_btn.ready()

    def _download_speech(self):
        def on_done(file: IO):
            try:
                self._play_audio(file.name)
                file.close()
            except Exception as exc:
                logging.error(exc)
                self._on_speech_failed()

        def on_fail(error: ProviderError):
            self._on_speech_failed(error)

        if not self.provider["tts"]:
            return

        if self.current_speech:
            lang: str = self.provider["tts"].denormalize_lang(self.current_speech["lang"])  # type: ignore
            self.provider["tts"].speech(self.current_speech["text"], lang, on_done, on_fail)

    def _on_speech_failed(self, error: ProviderError | None = None):
        text = _("Text-to-Speech failed")
        action: _NotificationAction | None = None

        if error and error.code == ProviderErrorCode.NETWORK:
            text = _("Text-to-Speech failed, check for network issues")

        if self.current_speech:
            called_from = self.current_speech["called_from"]
            action = {
                "label": _("Retry"),
                "name": "win.listen-src" if called_from == "src" else "win.listen-dest",
            }

            button_text = _("Text-to-Speech failed. Retry?")
            if called_from == "src":
                self.src_speech_btn.error(button_text)
            else:
                self.dest_speech_btn.error(button_text)

        self.send_notification(text, action=action)
        self._speech_reset(False)

    def _play_audio(self, path: str):
        if not self.player:
            return

        uri = "file://" + path
        self.player.set_property("uri", uri)
        self.player.set_state(Gst.State.PLAYING)
        self.add_tick_callback(self._gst_progress_timeout)

    def _on_gst_message(self, _bus, message: Gst.Message):
        if message.type == Gst.MessageType.EOS or message.type == Gst.MessageType.ERROR:
            if message.type == Gst.MessageType.ERROR:
                logging.error("Some error occurred while trying to play.")
            self._speech_reset()

    def _gst_progress_timeout(self, _widget, _clock):
        if not self.player:
            return False

        if self.current_speech and self.player.get_state(Gst.CLOCK_TIME_NONE) != Gst.State.NULL:
            have_pos, pos = self.player.query_position(Gst.Format.TIME)
            have_dur, dur = self.player.query_duration(Gst.Format.TIME)

            if have_pos and have_dur:
                if self.current_speech["called_from"] == "src":
                    self.src_speech_btn.progress(pos / dur)
                else:
                    self.dest_speech_btn.progress(pos / dur)

                if self.speech_loading:
                    self.speech_loading = False
                    self._check_speech_enabled()

            return True

        return False

    @Gtk.Template.Callback()
    def _on_key_event(self, _ctrl, keyval: int, _keycode: int, state: Gdk.ModifierType):
        """Called on self.win_key_ctrlr::key-pressed signal"""
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        shift_mask = Gdk.ModifierType.SHIFT_MASK
        unicode_key_val = Gdk.keyval_to_unicode(keyval)
        if (
            GLib.unichar_isgraph(chr(unicode_key_val))
            and modifiers in (shift_mask, 0)
            and not self.dest_text.props.editable
            and not self.src_text.is_focus()
        ):
            self.src_text.grab_focus()
            end_iter = self.src_buffer.get_end_iter()
            self.src_buffer.insert(end_iter, chr(unicode_key_val))
            return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE

    @Gtk.Template.Callback()
    def _on_src_activated(self, _texview):
        """Called on self.src_text::active signal"""
        if not Settings.get().live_translation:
            self.translation()

    @Gtk.Template.Callback()
    def _on_mistakes_clicked(self, _button, _data):
        """Called on self.mistakes_label::activate-link signal"""
        self.mistakes.props.reveal_child = False
        if self.trans_mistakes[1]:
            self.src_buffer.props.text = self.trans_mistakes[1]
        # Run translation again
        self.translation()

        return Gdk.EVENT_STOP

    def on_src_text_changed(self, buffer: Gtk.TextBuffer):
        if not self.provider["trans"]:
            return

        char_count = buffer.get_char_count()

        # If the text is over the highest number of characters allowed, it is truncated.
        # This is done for avoiding exceeding the limit imposed by translation services.
        if self.provider["trans"].chars_limit == -1:  # -1 means unlimited
            self.char_counter.props.label = ""
        else:
            self.char_counter.props.label = f'{str(char_count)}/{self.provider["trans"].chars_limit}'

            if char_count >= self.provider["trans"].chars_limit:
                self.send_notification(_("{} characters limit reached!").format(self.provider["trans"].chars_limit))
                buffer.delete(buffer.get_iter_at_offset(self.provider["trans"].chars_limit), buffer.get_end_iter())

        sensitive = char_count != 0
        self.lookup_action("translation").props.enabled = sensitive  # type: ignore
        self.lookup_action("clear").props.enabled = sensitive  # type: ignore
        self._check_speech_enabled()

    def on_dest_text_changed(self, buffer: Gtk.TextBuffer):
        if not self.provider["trans"]:
            return

        sensitive = buffer.get_char_count() != 0
        self.lookup_action("copy").props.enabled = sensitive  # type: ignore
        self.lookup_action("suggest").set_enabled(  # type: ignore
            ProviderFeature.SUGGESTIONS in self.provider["trans"].features and sensitive
        )
        self._check_speech_enabled()

    def user_action_ended(self, _buffer):
        if Settings.get().live_translation:
            self.translation()

    # The history part
    def reset_return_forward_btns(self):
        self.lookup_action("back").props.enabled = self.current_history < len(self.provider["trans"].history) - 1  # type: ignore
        self.lookup_action("forward").props.enabled = self.current_history > 0  # type: ignore

    # Retrieve translation history
    def history_update(self):
        if not self.provider["trans"]:
            return

        self.reset_return_forward_btns()
        translation = self.provider["trans"].history[self.current_history]
        self.src_lang_selector.selected = translation.original[1]
        self.dest_lang_selector.selected = translation.original[2]
        self.src_buffer.props.text = translation.original[0]
        self.dest_buffer.props.text = translation.text

    def appeared_before(self):
        if not self.provider["trans"]:
            return

        src_language = self.src_lang_selector.selected
        dest_language = self.dest_lang_selector.selected
        src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
        if (
            len(self.provider["trans"].history) >= self.current_history + 1
            and (self.provider["trans"].history[self.current_history].original[1] == src_language or "auto")
            and self.provider["trans"].history[self.current_history].original[2] == dest_language
            and self.provider["trans"].history[self.current_history].original[0] == src_text
        ):
            return True
        return False

    @Gtk.Template.Callback()
    def translation(self, _action=None, _param=None):
        if not self.provider["trans"]:
            return

        # If it's like the last translation then it's useless to continue
        if not self.appeared_before():
            src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
            src_language = self.src_lang_selector.selected
            dest_language = self.dest_lang_selector.selected

            if self.ongoing_trans:
                self.next_trans = {"text": src_text, "src": src_language, "dest": dest_language}
                return

            if self.next_trans:
                src_text = self.next_trans["text"]
                src_language = self.next_trans["src"]
                dest_language = self.next_trans["dest"]
                self.next_trans = {}

            # Show feedback for start of translation.
            self.translation_loading()

            # If the two languages are the same, nothing is done
            if src_language != dest_language:
                if src_text != "":
                    self.ongoing_trans = True

                    src, dest = self.provider["trans"].denormalize_lang(src_language, dest_language)
                    self.provider["trans"].translate(
                        src_text, src, dest, self.on_translation_success, self.on_translation_fail
                    )
                else:
                    self.trans_mistakes = (None, None)
                    self.trans_src_pron = None
                    self.trans_dest_pron = None
                    self.dest_buffer.props.text = ""

                    if not self.ongoing_trans:
                        self.translation_finish()

    def on_translation_success(self, translation: Translation):
        if not self.provider["trans"]:
            return

        self.trans_warning.props.visible = False

        if translation.detected and self.src_lang_selector.selected == "auto":
            if Settings.get().src_auto:
                self.src_lang_selector.set_insight(self.provider["trans"].normalize_lang_code(translation.detected))
            else:
                self.src_lang_selector.selected = translation.detected

        self.dest_buffer.props.text = translation.text

        self.trans_mistakes = translation.mistakes
        self.trans_src_pron = translation.pronunciation[0]
        self.trans_dest_pron = translation.pronunciation[1]

        # Finally, translation is saved in history
        self.add_history_entry(translation)

        # Mistakes
        if ProviderFeature.MISTAKES in self.provider["trans"].features and not self.trans_mistakes == (None, None):
            self.mistakes_label.set_markup(_("Did you mean: ") + f'<a href="#">{self.trans_mistakes[0]}</a>')
            self.mistakes.props.reveal_child = True
        elif self.mistakes.props.reveal_child:
            self.mistakes.props.reveal_child = False

        # Pronunciation
        reveal = Settings.get().show_pronunciation
        if ProviderFeature.PRONUNCIATION in self.provider["trans"].features:
            if self.trans_src_pron is not None and self.trans_mistakes == (None, None):
                self.src_pron_label.props.label = self.trans_src_pron
                self.src_pron_revealer.props.reveal_child = reveal
            elif self.src_pron_revealer.props.reveal_child:
                self.src_pron_revealer.props.reveal_child = False

            if self.trans_dest_pron is not None:
                self.dest_pron_label.props.label = self.trans_dest_pron
                self.dest_pron_revealer.props.reveal_child = reveal
            elif self.dest_pron_revealer.props.reveal_child:
                self.dest_pron_revealer.props.reveal_child = False

        self.ongoing_trans = False
        if self.next_trans:
            self.translation()
        else:
            self.translation_finish()

    def on_translation_fail(self, error: ProviderError):
        if not self.next_trans:
            self.translation_finish()
        self.trans_warning.props.visible = True
        self.ongoing_trans = False

        match error.code:
            case ProviderErrorCode.NETWORK:
                self.send_notification(
                    _("Translation failed, check for network issues"),
                    action={
                        "label": _("Retry"),
                        "name": "win.translation",
                    },
                )
            case ProviderErrorCode.API_KEY_INVALID:
                self.send_notification(
                    _("The provided API key is invalid"),
                    action={
                        "label": _("Retry"),
                        "name": "win.translation",
                    },
                )
            case ProviderErrorCode.API_KEY_REQUIRED:
                self.send_notification(
                    _("API key is required to use the service"),
                    action={
                        "label": _("Preferences"),
                        "name": "app.preferences",
                    },
                )
            case _:
                self.send_notification(
                    _("Translation failed"),
                    action={
                        "label": _("Retry"),
                        "name": "win.translation",
                    },
                )

    def translation_loading(self):
        self.trans_spinner.show()
        self.dest_box.props.sensitive = False
        self.langs_button_box.props.sensitive = False

    def translation_finish(self):
        self.trans_spinner.hide()
        self.dest_box.props.sensitive = True
        self.langs_button_box.props.sensitive = True

    def reload_translator(self):
        self.translator_loading = True

        # Load translator
        self.load_translator()

    def _on_active_provider_changed(self, _settings: Gio.Settings, _provider: str, kind: str):
        self.save_settings()
        match kind:
            case "trans":
                self.reload_translator()
            case "tts":
                self.load_tts()

    def _on_provider_changed(self, _settings: Gio.Settings, _key: str, name: str):
        if not self.translator_loading:
            if self.provider["trans"] and name == self.provider["trans"].name:
                self.reload_translator()

            if self.provider["tts"] and name == self.provider["tts"].name:
                self.load_tts()



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable

from gi.repository import Gio, GObject

from dialect.define import LANGUAGES


def get_lang_name(code: str) -> str | None:
    name = LANGUAGES.get(code)
    if name:
        name = gettext(name)
    return name


class LangObject(GObject.Object):
    __gtype_name__ = "LangObject"

    code: str = GObject.Property(type=str)  # type: ignore
    name: str = GObject.Property(type=str)  # type: ignore
    selected: bool = GObject.Property(type=bool, default=False)  # type: ignore

    def __init__(self, code: str, name: str, selected=False):
        super().__init__()

        self.code = code
        self.name = name
        self.selected = selected

    def __str__(self) -> str:
        return self.code


class LanguagesListModel(GObject.GObject, Gio.ListModel):
    __gtype_name__ = "LanguagesListModel"

    def __init__(self, names_func: Callable[[str], str | None] = get_lang_name):
        super().__init__()

        self.names_func = names_func
        self.langs: list[LangObject] = []

    def __iter__(self):
        return iter(self.langs)

    def do_get_item(self, position: int) -> LangObject:
        return self.langs[position]

    def do_get_item_type(self):
        return LangObject

    def do_get_n_items(self) -> int:
        return len(self.langs)

    def set_langs(self, langs: list[str], auto=False):
        removed = len(self.langs)
        self.langs.clear()

        if auto:
            self.langs.append(LangObject("auto", _("Auto")))

        for code in langs:
            self.langs.append(LangObject(code, self.names_func(code) or code))

        self.items_changed(0, removed, len(self.langs))

    def set_selected(self, code: str):
        for item in self.langs:
            item.props.selected = item.code == code



# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import re

from bs4 import BeautifulSoup, Tag

from dialect.providers.base import (
    ProviderCapability,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
    Translation,
)
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "bing"
    prettyname = "Bing"

    capabilities = ProviderCapability.TRANSLATION
    features = ProviderFeature.DETECTION | ProviderFeature.PRONUNCIATION

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    _headers = {"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)", "Accept": "*/*"}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 1000  # Web UI limit

        # Session vars
        self._key = ""
        self._token = ""
        self._ig = ""
        self._iid = ""
        self._count = 1

    @property
    def html_url(self):
        return self.format_url("www.bing.com", "/translator")

    @property
    def translate_url(self):
        params = {
            "isVertical": "1",
            "": "",
            "IG": self._ig,
            "IID": f"{self._iid}.{self._count}",
        }
        return self.format_url("www.bing.com", "/ttranslatev3", params)

    def init_trans(self, on_done, on_fail):
        def on_response(data):
            if data:
                try:
                    soup = BeautifulSoup(data, "html.parser")

                    # Get Langs
                    langs = soup.find("optgroup", {"id": "t_tgtAllLang"})
                    if isinstance(langs, Tag):
                        for child in langs.findChildren():
                            if child.name == "option":
                                self.add_lang(child["value"], child.contents[0])

                    # Get IID
                    iid = soup.find("div", {"id": "rich_tta"})
                    if isinstance(iid, Tag):
                        self._iid = iid["data-iid"]

                    # Decode response bytes
                    data = data.decode("utf-8")

                    # Look for abuse prevention data
                    params = re.findall(r"var params_AbusePreventionHelper = \[(.*?)\];", data)[0]  # noqa
                    abuse_params = params.replace('"', "").split(",")
                    self._key = abuse_params[0]
                    self._token = abuse_params[1]

                    # Look for IG
                    self._ig = re.findall('IG:"(.*?)",', data)[0]

                    on_done()

                except Exception as exc:
                    error = "Failed parsing HTML from bing.com"
                    logging.warning(error, exc)
                    on_fail(ProviderError(ProviderErrorCode.NETWORK, error))

            else:
                on_fail(ProviderError(ProviderErrorCode.EMPTY, "Could not get HTML from bing.com"))

        # Message request to get bing's website html
        message = self.create_message("GET", self.html_url, headers=self._headers)

        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail, False, False)

    def translate(self, text, src, dest, on_done, on_fail):
        def on_response(data):
            try:
                data = data[0]
                detected = None
                pronunciation = None

                if "translations" in data:
                    if "detectedLanguage" in data:
                        detected = data["detectedLanguage"]["language"]

                    if "transliteration" in data["translations"][0]:
                        pronunciation = data["translations"][0]["transliteration"]["text"]

                    translation = Translation(
                        data["translations"][0]["text"],
                        (text, src, dest),
                        detected=detected,
                        pronunciation=(None, pronunciation),
                    )
                    on_done(translation)

            except Exception as exc:
                logging.warning(exc)
                on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, str(exc)))

        # Increment requests count
        self._count += 1

        # Form data
        data = {
            "fromLang": "auto-detect" if src == "auto" else src,
            "text": text,
            "to": dest,
            "token": self._token,
            "key": self._key,
        }
        # Request message
        message = self.create_message("POST", self.translate_url, data, self._headers, True)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def check_known_errors(self, _status, data):
        if not data:
            return ProviderError(ProviderErrorCode.EMPTY, "Response is empty!")

        if "errorMessage" in data:
            error = data["errorMessage"]
            code = data["statusCode"]

            match code:
                case _:
                    return ProviderError(ProviderErrorCode.UNEXPECTED, error)

        return None



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import urllib.parse
from dataclasses import dataclass
from enum import Enum, Flag, auto
from typing import IO, Callable

from dialect.define import LANG_ALIASES
from dialect.languages import get_lang_name
from dialect.providers.settings import ProviderDefaults, ProviderSettings


class ProviderCapability(Flag):
    TRANSLATION = auto()
    """ If it provides translation """
    TTS = auto()
    """ If it provides text-to-speech """
    DEFINITIONS = auto()
    """ If it provides dictionary definitions """


class ProviderFeature(Flag):
    NONE = auto()
    """ Provider has no features """
    INSTANCES = auto()
    """ If it supports changing the instance url """
    API_KEY = auto()
    """ If the api key is supported but not necessary """
    API_KEY_REQUIRED = auto()
    """ If the api key is required for the provider to work """
    API_KEY_USAGE = auto()
    """ If the service reports api usage """
    DETECTION = auto()
    """ If it supports detecting text language (Auto translation) """
    MISTAKES = auto()
    """ If it supports showing translation mistakes """
    PRONUNCIATION = auto()
    """ If it supports showing translation pronunciation """
    SUGGESTIONS = auto()
    """ If it supports sending translation suggestions to the service """


class ProvideLangModel(Enum):
    STATIC = auto()
    """
    The provider populate its `src_languages` and `dest_languages` properties.
    The `cmp_langs` method will be used to decide if one code can be translated to another.
    """
    DYNAMIC = auto()
    """
    The provider only populate its `src_languages` property.
    The `dest_langs_for` method will be used to get possible destination codes for a code.
    """


class ProviderErrorCode(Enum):
    UNEXPECTED = auto()
    NETWORK = auto()
    EMPTY = auto()
    API_KEY_REQUIRED = auto()
    API_KEY_INVALID = auto()
    INVALID_LANG_CODE = auto()
    BATCH_SIZE_EXCEEDED = auto()
    CHARACTERS_LIMIT_EXCEEDED = auto()
    SERVICE_LIMIT_REACHED = auto()
    TRANSLATION_FAILED = auto()
    TTS_FAILED = auto()


class ProviderError:
    """Helper error handing class to be passed between callbacks"""

    def __init__(self, code: ProviderErrorCode, message: str = "") -> None:
        self.code = code  # Serves for quick error matching
        self.message = message  # More detailed error info if needed


@dataclass
class Translation:
    text: str
    original: tuple[str, str, str]
    detected: str | None = None
    mistakes: tuple[str | None, str | None] = (None, None)
    pronunciation: tuple[str | None, str | None] = (None, None)


class BaseProvider:
    name = ""
    """ Module name for code use, like settings storing """
    prettyname = ""
    """ Module name for UI display """
    capabilities: ProviderCapability | None = None
    """ Provider capabilities, translation, tts, etc """
    features: ProviderFeature = ProviderFeature.NONE
    """ Provider features """
    lang_model: ProvideLangModel = ProvideLangModel.STATIC
    """ Translation language model """

    defaults: ProviderDefaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }
    """ Default provider settings """

    def __init__(self):
        self.src_languages: list[str] = []
        """ Source languages available for translating """
        self.dest_languages: list[str] = []
        """ Destination languages available for translating """
        self.tts_languages: list[str] = []
        """ Languages available for TTS """
        self._nonstandard_langs: dict[str, str] = {}
        """ Mapping of lang codes that differ with Dialect ones """
        self._languages_names: dict[str, str] = {}
        """ Names of languages provided by the service """

        self.chars_limit: int = -1
        """ Translation char limit """

        self.history: list[Translation] = []
        """ Here we save the translation history """

        # GSettings
        self.settings = ProviderSettings(self.name, self.defaults)

    """
    Providers API methods
    """

    def validate_instance(self, url: str, on_done: Callable[[bool], None], on_fail: Callable[[ProviderError], None]):
        """
        Validate an instance of the provider.

        Args:
            url: The instance URL to test, only hostname and tld, e.g. libretranslate.com, localhost
            on_done: Called when the validation is done, argument is the result of the validation
            on_fail: Called when there's a fail in the validation process
        """
        raise NotImplementedError()

    def validate_api_key(self, key: str, on_done: Callable[[bool], None], on_fail: Callable[[ProviderError], None]):
        """
        Validate an API key.

        Args:
            key: The API key to validate
            on_done: Called when the validation is done, argument is the result of the validation
            on_fail: Called when there's a fail in the validation process
        """
        raise NotImplementedError()

    def init_trans(self, on_done: Callable[[], None], on_fail: Callable[[ProviderError], None]):
        """
        Initializes the provider translation capabilities.

        Args:
            on_done: Called after the provider was successfully initialized
            on_fail: Called after any error on initialization
        """
        on_done()

    def init_tts(self, on_done: Callable[[], None], on_fail: Callable[[ProviderError], None]):
        """
        Initializes the provider text-to-speech capabilities.

        Args:
            on_done: Called after the provider was successfully initialized
            on_fail: Called after any error on initialization
        """
        on_done()

    def translate(
        self,
        text: str,
        src: str,
        dest: str,
        on_done: Callable[[Translation], None],
        on_fail: Callable[[ProviderError], None],
    ):
        """
        Translates text in the provider.

        Args:
            text: The text to translate
            src: The lang code of the source text
            dest: The lang code to translate the text to
            on_done: Called after the text was successfully translated
            on_fail: Called after any error on translation
        """
        raise NotImplementedError()

    def suggest(
        self,
        text: str,
        src: str,
        dest: str,
        suggestion: str,
        on_done: Callable[[bool], None],
        on_fail: Callable[[ProviderError], None],
    ):
        """
        Sends a translation suggestion to the provider.

        Args:
            text: Original text without translation
            src: The lang code of the original text
            dest: The lang code of the translated text
            suggestion: Suggested translation for text
            on_done: Called after the suggestion was successfully send, argument means if it was accepted or rejected
            on_fail: Called after any error on the suggestion process
        """
        raise NotImplementedError()

    def speech(
        self,
        text: str,
        language: str,
        on_done: Callable[[IO], None],
        on_fail: Callable[[ProviderError], None],
    ):
        """
        Generate speech audio from text

        Args:
            text: Text to generate speech from
            language: The lang code of text
            on_done: Called after the process successful
            on_fail: Called after any error on the speech process
        """
        raise NotImplementedError()

    def api_char_usage(
        self,
        on_done: Callable[[int, int], None],
        on_fail: Callable[[ProviderError], None],
    ):
        """
        Retrieves the API usage status

        Args:
            on_done: Called after the process successful, with the usage and limit as args
            on_fail: Called after any error on the speech process
        """
        raise NotImplementedError()

    def cmp_langs(self, a: str, b: str) -> bool:
        """
        Compare two language codes.

        It assumes that the codes have been normalized by `normalize_lang_code`.

        This method exists so providers can add additional comparison logic.

        Args:
            a: First lang to compare
            b: Second lang to compare
        """

        return a == b

    def dest_langs_for(self, code: str) -> list[str]:
        """
        Get the available destination languages for a source language.
        """
        raise NotImplementedError()

    @property
    def lang_aliases(self) -> dict[str, str]:
        """
        Mapping of Dialect/CLDR's lang codes to the provider ones.

        Some providers might use different lang codes from the ones used by Dialect to for example get localized
        language names.

        This dict is used by `add_lang` so lang codes can later be denormalized with `denormalize_lang`.

        Codes must be formatted with the criteria from normalize_lang_code, because this value would be used by
        `add_lang` after normalization.

        Check `dialect.define.LANG_ALIASES` for reference mappings.
        """
        return {}

    """
    Provider settings helpers and properties
    """

    @property
    def instance_url(self) -> str:
        """Instance url saved on settings."""
        return self.settings.instance_url

    @instance_url.setter
    def instance_url(self, url: str):
        self.settings.instance_url = url

    def reset_instance_url(self):
        """Resets saved instance url."""
        self.instance_url = ""

    @property
    def api_key(self) -> str:
        """API key saved on settings."""
        return self.settings.api_key

    @api_key.setter
    def api_key(self, api_key: str):
        self.settings.api_key = api_key

    def reset_api_key(self):
        """Resets saved API key."""
        self.api_key = ""

    @property
    def recent_src_langs(self) -> list[str]:
        """Saved recent source langs of the user."""
        return self.settings.src_langs

    @recent_src_langs.setter
    def recent_src_langs(self, src_langs: list[str]):
        self.settings.src_langs = src_langs

    def reset_src_langs(self):
        """Reset saved recent user source langs"""
        self.recent_src_langs = []

    @property
    def recent_dest_langs(self) -> list[str]:
        """Saved recent destination langs of the user."""
        return self.settings.dest_langs

    @recent_dest_langs.setter
    def recent_dest_langs(self, dest_langs: list[str]):
        self.settings.dest_langs = dest_langs

    def reset_dest_langs(self):
        """Reset saved recent user destination langs"""
        self.recent_dest_langs = []

    """
    General provider helpers
    """

    @staticmethod
    def format_url(url: str, path: str = "", params: dict = {}, http: bool = False):
        """
        Compose a HTTP url with the given pieces.

        If url is localhost, `http` is ignored and HTTP protocol is forced.

        Args:
            url: Base url, hostname and tld
            path: Path of the url
            params: Params to populate a url query
            http: If HTTP should be used instead of HTTPS
        """

        if not path.startswith("/"):
            path = "/" + path

        protocol = "https://"
        if url.startswith("localhost:") or http:
            protocol = "http://"

        params_str = urllib.parse.urlencode(params)
        if params_str:
            params_str = "?" + params_str

        return protocol + url + path + params_str

    def normalize_lang_code(self, code: str):
        """
        Normalice a language code to Dialect's criteria.

        Criteria:
        - Codes must be lowercase, e.g. ES => es
        - Codes can have a second code delimited by a hyphen, e.g. zh_CN => zh-CN
        - If second code is two chars long it's considered a country code and must be uppercase, e.g. zh-cn => zh-CN
        - If second code is four chars long it's considered a script code and must be capitalized,
        e.g. zh-HANS => zh-Hans

        This method also maps lang codes aliases using `lang_aliases` and `dialect.define.LANG_ALIASES`.

        Args:
            code: Language ISO code
        """
        code = code.replace("_", "-").lower()  # Normalize separator
        codes = code.split("-")

        if len(codes) == 2:  # Code contain a script or country code
            if len(codes[1]) == 4:  # ISO 15924 (script)
                codes[1] = codes[1].capitalize()

            elif len(codes[1]) == 2:  # ISO 3166-1 (country)
                codes[1] = codes[1].upper()

            code = "-".join(codes)

        aliases = {**LANG_ALIASES, **self.lang_aliases}
        if code in aliases:
            code = aliases[code]

        return code

    def add_lang(
        self,
        original_code: str,
        name: str | None = None,
        trans_src: bool = True,
        trans_dest: bool = True,
        tts: bool = False,
    ):
        """
        Add lang supported by provider after normalization.

        Normalized lang codes are saved for latter denormalization using `denormalize_lang`.

        Args:
            original_code: Lang code to add
            name: Language name to fallback in case Dialect doesn't provide one
            trans_src: Add language as supported for translation as src
            trans_dest: Add language as supported for translation as dest
            tts: Add language as supported for text-to-speech
        """

        code = self.normalize_lang_code(original_code)  # Get normalized lang code

        if trans_src:  # Add lang to supported languages list
            self.src_languages.append(code)
        if trans_dest:
            self.dest_languages.append(code)
        if tts:  # Add lang to supported TTS languages list
            self.tts_languages.append(code)

        if code != original_code and code not in self._nonstandard_langs:
            # Save a divergent lang code for later denormalization
            self._nonstandard_langs[code] = original_code

        if name is not None and code not in self._languages_names:
            # Save name provider by the service
            self._languages_names[code] = name

    def denormalize_lang(self, *codes: str) -> str | tuple[str, ...]:
        """
        Get denormalized lang code if available.

        This method will return a tuple with the same length of given codes or a str if only one code was passed.

        Args:
        *codes: Lang codes to denormalize
        """

        if len(codes) == 1:
            return self._nonstandard_langs.get(codes[0], codes[0])

        result = []
        for code in codes:
            result.append(self._nonstandard_langs.get(code, code))
        return tuple(result)

    def get_lang_name(self, code: str) -> str | None:
        """
        Get a localized language name.

        Fallback to a name provided by the provider if available or ultimately just the code.

        Args:
            code: Language to get a name for
        """
        name = get_lang_name(code)  # Try getting translated name from Dialect

        if name is None:  # Get name from provider if available
            return self._languages_names.get(code, code)

        return name



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from dialect.define import RES_PATH
from dialect.settings import Settings


@Gtk.Template(resource_path=f"{RES_PATH}/shortcuts.ui")
class DialectShortcutsWindow(Gtk.ShortcutsWindow):
    __gtype_name__ = "DialectShortcutsWindow"

    translate_shortcut: Gtk.ShortcutsShortcut = Gtk.Template.Child()  # type: ignore

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @Gtk.Template.Callback()
    def _on_show(self, _data):
        """Called on self::show signal"""
        self.translate_shortcut.props.visible = not Settings.get().live_translation
        self.translate_shortcut.props.accelerator = Settings.get().translate_accel



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from tempfile import NamedTemporaryFile
from urllib.parse import quote

from dialect.providers.base import (
    ProviderCapability,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
    Translation,
)
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "lingva"
    prettyname = "Lingva Translate"

    capabilities = ProviderCapability.TRANSLATION | ProviderCapability.TTS
    features = (
        ProviderFeature.INSTANCES | ProviderFeature.DETECTION | ProviderFeature.MISTAKES | ProviderFeature.PRONUNCIATION
    )

    defaults = {
        "instance_url": "lingva.dialectapp.org",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 5000

    def validate_instance(self, url, on_done, on_fail):
        def on_response(data):
            valid = False
            try:
                valid = "translation" in data
            except:  # noqa
                pass

            on_done(valid)

        # Lingva translation endpoint
        message = self.create_message("GET", self.format_url(url, "/api/v1/en/es/hello"))
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail, False)

    @property
    def lang_url(self):
        return self.format_url(self.instance_url, "/api/v1/languages/")

    @property
    def translate_url(self):
        return self.format_url(self.instance_url, "/api/v1/{src}/{dest}/{text}")

    @property
    def speech_url(self):
        return self.format_url(self.instance_url, "/api/v1/audio/{lang}/{text}")

    def init(self, on_done, on_fail):
        def on_response(data):
            if "languages" in data:
                for lang in data["languages"]:
                    if lang["code"] != "auto":
                        self.add_lang(lang["code"], lang["name"], tts=True)
                on_done()
            else:
                on_fail(ProviderError(ProviderErrorCode.UNEXPECTED, "No langs found in server."))

        # Languages message request
        message = self.create_message("GET", self.lang_url)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def init_trans(self, on_done, on_fail):
        self.init(on_done, on_fail)

    def init_tts(self, on_done, on_fail):
        self.init(on_done, on_fail)

    def translate(self, text, src, dest, on_done, on_fail):
        def on_response(data):
            try:
                detected = data.get("info", {}).get("detectedSource", None)
                mistakes = data.get("info", {}).get("typo", None)
                src_pronunciation = data.get("info", {}).get("pronunciation", {}).get("query", None)
                dest_pronunciation = data.get("info", {}).get("pronunciation", {}).get("translation", None)

                translation = Translation(
                    data["translation"],
                    (text, src, dest),
                    detected,
                    (mistakes, mistakes),
                    (src_pronunciation, dest_pronunciation),
                )

                on_done(translation)

            except Exception as exc:
                error = "Failed reading the translation data"
                logging.warning(error, exc)
                on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, error))

        # Format url query data
        text = quote(text, safe="")
        url = self.translate_url.format(text=text, src=src, dest=dest)

        # Request message
        message = self.create_message("GET", url)

        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def speech(self, text, language, on_done, on_fail):
        def on_response(data):
            if "audio" in data:
                file = NamedTemporaryFile()
                audio = bytearray(data["audio"])
                file.write(audio)
                file.seek(0)

                on_done(file)
            else:
                on_fail(ProviderError(ProviderErrorCode.TTS_FAILED, "No audio was found."))

        # Format url query data
        url = self.speech_url.format(text=text, lang=language)

        # Request message
        message = self.create_message("GET", url)

        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def check_known_errors(self, _status, data):
        """Raises a proper Exception if an error is found in the data."""
        if not data:
            return ProviderError(ProviderErrorCode.EMPTY, "Response is empty!")
        if "error" in data:
            error = data["error"]

            if error == "Invalid target language" or error == "Invalid source language":
                return ProviderError(ProviderErrorCode.INVALID_LANG_CODE, error)
            else:
                return ProviderError(ProviderErrorCode.UNEXPECTED, error)



# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from uuid import uuid4

from dialect.providers.base import (
    ProviderCapability,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
    Translation,
)
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "yandex"
    prettyname = "Yandex"

    capabilities = ProviderCapability.TRANSLATION
    features = ProviderFeature.DETECTION

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    _headers = {
        "User-Agent": "ru.yandex.translate/21.15.4.21402814 (Xiaomi Redmi K20 Pro; Android 11)",
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 10000

        self._uuid = str(uuid4()).replace("-", "")

    def init_trans(self, on_done, on_fail):
        languages = [
            "af",
            "sq",
            "am",
            "ar",
            "hy",
            "az",
            "ba",
            "eu",
            "be",
            "bn",
            "bs",
            "bg",
            "my",
            "ca",
            "ceb",
            "zh",
            "cv",
            "hr",
            "cs",
            "da",
            "nl",
            "sjn",
            "emj",
            "en",
            "eo",
            "et",
            "fi",
            "fr",
            "gl",
            "ka",
            "de",
            "el",
            "gu",
            "ht",
            "he",
            "mrj",
            "hi",
            "hu",
            "is",
            "id",
            "ga",
            "it",
            "ja",
            "jv",
            "kn",
            "kk",
            "kazlat",
            "km",
            "ko",
            "ky",
            "lo",
            "la",
            "lv",
            "lt",
            "lb",
            "mk",
            "mg",
            "ms",
            "ml",
            "mt",
            "mi",
            "mr",
            "mhr",
            "mn",
            "ne",
            "no",
            "pap",
            "fa",
            "pl",
            "pt",
            "pa",
            "ro",
            "ru",
            "gd",
            "sr",
            "si",
            "sk",
            "sl",
            "es",
            "su",
            "sw",
            "sv",
            "tl",
            "tg",
            "ta",
            "tt",
            "te",
            "th",
            "tr",
            "udm",
            "uk",
            "ur",
            "uz",
            "uzbcyr",
            "vi",
            "cy",
            "xh",
            "sah",
            "yi",
            "zu",
        ]
        for code in languages:
            self.add_lang(code)

        on_done()

    @property
    def translate_url(self):
        path = f"/api/v1/tr.json/translate?id={self._uuid}-0-0&srv=android"
        return self.format_url("translate.yandex.net", path)

    def translate(self, text, src, dest, on_done, on_fail):
        def on_response(data):
            try:
                detected = None
                if "code" in data and data["code"] == 200:
                    if "lang" in data:
                        detected = data["lang"].split("-")[0]

                    if "text" in data:
                        translation = Translation(data["text"][0], (text, src, dest), detected)
                        on_done(translation)

                    else:
                        on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, "Translation failed"))

                else:
                    error = data["message"] if "message" in data else ""
                    on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, error))
            except Exception as exc:
                error = "Failed reading the translation data"
                logging.warning(error, exc)
                on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, error))

        # Form data
        data = {"lang": dest, "text": text}
        if src != "auto":
            data["lang"] = f"{src}-{dest}"

        # Request message
        message = self.create_message("POST", self.translate_url, data, self._headers, True)

        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import html
import json
import logging
import random
import re
from tempfile import NamedTemporaryFile

from gtts import gTTS, lang

from dialect.providers.base import (
    ProviderCapability,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
    Translation,
)
from dialect.providers.local import LocalProvider
from dialect.providers.soup import SoupProvider

RPC_ID = "MkEWBc"

# Predefined URLs used to make google translate requests.
TRANSLATE_RPC = "{host}/_/TranslateWebserverUi/data/batchexecute"

DEFAULT_SERVICE_URLS = (
    "translate.google.ac",
    "translate.google.ad",
    "translate.google.ae",
    "translate.google.al",
    "translate.google.am",
    "translate.google.as",
    "translate.google.at",
    "translate.google.az",
    "translate.google.ba",
    "translate.google.be",
    "translate.google.bf",
    "translate.google.bg",
    "translate.google.bi",
    "translate.google.bj",
    "translate.google.bs",
    "translate.google.bt",
    "translate.google.by",
    "translate.google.ca",
    "translate.google.cat",
    "translate.google.cc",
    "translate.google.cd",
    "translate.google.cf",
    "translate.google.cg",
    "translate.google.ch",
    "translate.google.ci",
    "translate.google.cl",
    "translate.google.cm",
    "translate.google.cn",
    "translate.google.co.ao",
    "translate.google.co.bw",
    "translate.google.co.ck",
    "translate.google.co.cr",
    "translate.google.co.id",
    "translate.google.co.il",
    "translate.google.co.in",
    "translate.google.co.jp",
    "translate.google.co.ke",
    "translate.google.co.kr",
    "translate.google.co.ls",
    "translate.google.co.ma",
    "translate.google.co.mz",
    "translate.google.co.nz",
    "translate.google.co.th",
    "translate.google.co.tz",
    "translate.google.co.ug",
    "translate.google.co.uk",
    "translate.google.co.uz",
    "translate.google.co.ve",
    "translate.google.co.vi",
    "translate.google.co.za",
    "translate.google.co.zm",
    "translate.google.co.zw",
    "translate.google.com.af",
    "translate.google.com.ag",
    "translate.google.com.ai",
    "translate.google.com.ar",
    "translate.google.com.au",
    "translate.google.com.bd",
    "translate.google.com.bh",
    "translate.google.com.bn",
    "translate.google.com.bo",
    "translate.google.com.br",
    "translate.google.com.bz",
    "translate.google.com.co",
    "translate.google.com.cu",
    "translate.google.com.cy",
    "translate.google.com.do",
    "translate.google.com.ec",
    "translate.google.com.eg",
    "translate.google.com.et",
    "translate.google.com.fj",
    "translate.google.com.gh",
    "translate.google.com.gi",
    "translate.google.com.gt",
    "translate.google.com.hk",
    "translate.google.com.jm",
    "translate.google.com.kh",
    "translate.google.com.kw",
    "translate.google.com.lb",
    "translate.google.com.ly",
    "translate.google.com.mm",
    "translate.google.com.mt",
    "translate.google.com.mx",
    "translate.google.com.my",
    "translate.google.com.na",
    "translate.google.com.ng",
    "translate.google.com.ni",
    "translate.google.com.np",
    "translate.google.com.om",
    "translate.google.com.pa",
    "translate.google.com.pe",
    "translate.google.com.pg",
    "translate.google.com.ph",
    "translate.google.com.pk",
    "translate.google.com.pr",
    "translate.google.com.py",
    "translate.google.com.qa",
    "translate.google.com.sa",
    "translate.google.com.sb",
    "translate.google.com.sg",
    "translate.google.com.sl",
    "translate.google.com.sv",
    "translate.google.com.tj",
    "translate.google.com.tr",
    "translate.google.com.tw",
    "translate.google.com.ua",
    "translate.google.com.uy",
    "translate.google.com.vc",
    "translate.google.com.vn",
    "translate.google.com",
    "translate.google.cv",
    "translate.google.cz",
    "translate.google.de",
    "translate.google.dj",
    "translate.google.dk",
    "translate.google.dm",
    "translate.google.dz",
    "translate.google.ee",
    "translate.google.es",
    "translate.google.fi",
    "translate.google.fm",
    "translate.google.fr",
    "translate.google.ga",
    "translate.google.ge",
    "translate.google.gg",
    "translate.google.gl",
    "translate.google.gm",
    "translate.google.gp",
    "translate.google.gr",
    "translate.google.gy",
    "translate.google.hn",
    "translate.google.hr",
    "translate.google.ht",
    "translate.google.hu",
    "translate.google.ie",
    "translate.google.im",
    "translate.google.iq",
    "translate.google.is",
    "translate.google.it",
    "translate.google.je",
    "translate.google.jo",
    "translate.google.kg",
    "translate.google.ki",
    "translate.google.kz",
    "translate.google.la",
    "translate.google.li",
    "translate.google.lk",
    "translate.google.lt",
    "translate.google.lu",
    "translate.google.lv",
    "translate.google.md",
    "translate.google.me",
    "translate.google.mg",
    "translate.google.mk",
    "translate.google.ml",
    "translate.google.mn",
    "translate.google.ms",
    "translate.google.mu",
    "translate.google.mv",
    "translate.google.mw",
    "translate.google.ne",
    "translate.google.nl",
    "translate.google.no",
    "translate.google.nr",
    "translate.google.nu",
    "translate.google.pl",
    "translate.google.pn",
    "translate.google.ps",
    "translate.google.pt",
    "translate.google.ro",
    "translate.google.rs",
    "translate.google.ru",
    "translate.google.rw",
    "translate.google.sc",
    "translate.google.se",
    "translate.google.sh",
    "translate.google.si",
    "translate.google.sk",
    "translate.google.sm",
    "translate.google.sn",
    "translate.google.so",
    "translate.google.sr",
    "translate.google.st",
    "translate.google.td",
    "translate.google.tg",
    "translate.google.tk",
    "translate.google.tl",
    "translate.google.tm",
    "translate.google.tn",
    "translate.google.to",
    "translate.google.tt",
    "translate.google.us",
    "translate.google.vg",
    "translate.google.vu",
    "translate.google.ws",
)


class Provider(LocalProvider, SoupProvider):
    name = "google"
    prettyname = "Google"

    capabilities = ProviderCapability.TRANSLATION | ProviderCapability.TTS
    features = ProviderFeature.DETECTION | ProviderFeature.MISTAKES | ProviderFeature.PRONUNCIATION

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    _service_urls = DEFAULT_SERVICE_URLS
    _headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)",
        "Referer": "https://translate.google.com",
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 2000

    def init_trans(self, on_done, on_fail):
        languages = [
            "af",
            "sq",
            "am",
            "ar",
            "hy",
            "az",
            "eu",
            "be",
            "bn",
            "bs",
            "bg",
            "ca",
            "ceb",
            "ny",
            "zh-CN",
            "zh-TW",
            "co",
            "hr",
            "cs",
            "da",
            "nl",
            "en",
            "eo",
            "et",
            "tl",
            "fi",
            "fr",
            "fy",
            "gl",
            "ka",
            "de",
            "el",
            "gu",
            "ht",
            "ha",
            "haw",
            "iw",
            "hi",
            "hmn",
            "hu",
            "is",
            "ig",
            "id",
            "ga",
            "it",
            "ja",
            "jw",
            "kn",
            "kk",
            "km",
            "rw",
            "ko",
            "ku",
            "ky",
            "lo",
            "la",
            "lv",
            "lt",
            "lb",
            "mk",
            "mg",
            "ms",
            "ml",
            "mt",
            "mi",
            "mr",
            "mn",
            "my",
            "ne",
            "no",
            "or",
            "ps",
            "fa",
            "pl",
            "pt",
            "pa",
            "ro",
            "ru",
            "sm",
            "gd",
            "sr",
            "st",
            "sn",
            "sd",
            "si",
            "sk",
            "sl",
            "so",
            "es",
            "su",
            "sw",
            "sv",
            "tg",
            "ta",
            "tt",
            "te",
            "th",
            "tr",
            "tk",
            "uk",
            "ur",
            "ug",
            "uz",
            "vi",
            "cy",
            "xh",
            "yi",
            "yo",
            "zu",
        ]
        for code in languages:
            self.add_lang(code)

        on_done()

    def init_tts(self, on_done, on_fail):
        for code in lang.tts_langs().keys():
            self.add_lang(code, trans_src=False, trans_dest=False, tts=True)

        on_done()

    @staticmethod
    def _build_rpc_request(text: str, src: str, dest: str):
        return json.dumps(
            [
                [
                    [
                        RPC_ID,
                        json.dumps([[text, src, dest, True], [None]], separators=(",", ":")),
                        None,
                        "generic",
                    ],
                ]
            ],
            separators=(",", ":"),
        )

    def _pick_service_url(self):
        if len(self._service_urls) == 1:
            return self._service_urls[0]
        return random.choice(self._service_urls)

    @property
    def translate_url(self):
        url = TRANSLATE_RPC.format(host=self._pick_service_url()) + "?"
        params = {
            "rpcids": RPC_ID,
            "bl": "boq_translate-webserver_20201207.13_p0",
            "soc-app": "1",
            "soc-platform": "1",
            "soc-device": "1",
            "rt": "c",
        }

        return self.format_url(url, params=params)

    def translate(self, text, src_lang, dest_lang, on_done, on_fail):
        def on_response(data):
            try:
                token_found = False
                square_bracket_counts = [0, 0]
                resp = ""
                data = data.decode("utf-8")

                for line in data.split("\n"):
                    token_found = token_found or f'"{RPC_ID}"' in line[:30]
                    if not token_found:
                        continue

                    is_in_string = False
                    for index, char in enumerate(line):
                        if char == '"' and line[max(0, index - 1)] != "\\":
                            is_in_string = not is_in_string
                        if not is_in_string:
                            if char == "[":
                                square_bracket_counts[0] += 1
                            elif char == "]":
                                square_bracket_counts[1] += 1

                    resp += line
                    if square_bracket_counts[0] == square_bracket_counts[1]:
                        break

                data = json.loads(resp)
                parsed = json.loads(data[0][2])
                translated_parts = None
                translated = None
                try:
                    translated_parts = list(
                        map(
                            lambda part: TranslatedPart(
                                part[0] if len(part) > 0 else "", part[1] if len(part) >= 2 else []
                            ),
                            parsed[1][0][0][5],
                        )
                    )
                except TypeError:
                    translated_parts = [TranslatedPart(parsed[1][0][1][0], [parsed[1][0][0][0], parsed[1][0][1][0]])]

                first_iter = True
                translated = ""
                for part in translated_parts:
                    if not part.text.isspace() and not first_iter:
                        translated += " "
                    if first_iter:
                        first_iter = False
                    translated += part.text

                src = None
                try:
                    src = parsed[1][-1][1]
                except (IndexError, TypeError):
                    pass

                if not src == src_lang:
                    on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, "source language mismatch"))
                    return

                if src == "auto":
                    try:
                        if parsed[0][2] in self.src_languages:
                            src = parsed[0][2]
                    except (IndexError, TypeError):
                        pass

                dest = None
                try:
                    dest = parsed[1][-1][2]
                except (IndexError, TypeError):
                    pass

                if not dest == dest_lang:
                    on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, "destination language mismatch"))
                    return

                origin_pronunciation = None
                try:
                    origin_pronunciation = parsed[0][0]
                except (IndexError, TypeError):
                    pass

                pronunciation = None
                try:
                    pronunciation = parsed[1][0][0][1]
                except (IndexError, TypeError):
                    pass

                mistake = None
                try:
                    mistake = parsed[0][1][0][0][1]
                    # Convert to pango markup
                    mistake = mistake.replace("<em>", "<b>").replace("</em>", "</b>")
                except (IndexError, TypeError):
                    pass

                result = Translation(
                    translated,
                    (text, src_lang, dest_lang),
                    src,
                    (mistake, self._strip_html_tags(mistake)),
                    (origin_pronunciation, pronunciation),
                )
                on_done(result)

            except Exception as exc:
                logging.warning(exc)
                on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, str(exc)))

        # Form data
        data = {
            "f.req": self._build_rpc_request(text, src_lang, dest_lang),
        }

        # Request message
        message = self.create_message("POST", self.translate_url, data, self._headers, True)

        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail, False, False)

    def _strip_html_tags(self, text):
        """Strip html tags"""
        if text is None:
            return None

        tags_re = re.compile(r"(<!--.*?-->|<[^>]*>)")
        tags_removed = tags_re.sub("", text)
        escaped = html.escape(tags_removed)
        return escaped

    def speech(self, text, language, on_done, on_fail):
        def work():
            try:
                file = NamedTemporaryFile()
                tts = gTTS(text, lang=language, lang_check=False)
                tts.write_to_fp(file)
                file.seek(0)

                on_done(file)
            except Exception as exc:
                logging.warning(exc)
                on_fail(ProviderError(ProviderErrorCode.TTS_FAILED, str(exc)))

        self.launch_thread(work)


class TranslatedPart:
    def __init__(self, text: str, candidates: list[str]):
        self.text = text
        self.candidates = candidates

    def __str__(self):
        return self.text

    def __dict__(self):
        return {
            "text": self.text,
            "candidates": self.candidates,
        }



# Copyright 2024 Mufeed Ali
# Copyright 2024 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from dialect.providers.base import (
    ProviderCapability,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
    Translation,
)
from dialect.providers.soup import SoupProvider

API_V = "v2"


class Provider(SoupProvider):
    name = "deepl"
    prettyname = "DeepL"

    capabilities = ProviderCapability.TRANSLATION
    features = (
        ProviderFeature.DETECTION
        | ProviderFeature.API_KEY
        | ProviderFeature.API_KEY_REQUIRED
        | ProviderFeature.API_KEY_USAGE
    )

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en-US"],
    }

    _api_free = "api-free.deepl.com"
    _api_pro = "api.deepl.com"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 5000

        # DeepL API Free keys can be identified by the suffix ":fx"
        self.api_url = self.__get_api_url(self.api_key)

    def __get_api_url(self, api_key: str) -> str:
        return self._api_free if api_key.endswith(":fx") else self._api_pro

    @property
    def source_lang_url(self):
        return self.format_url(self.api_url, f"/{API_V}/languages", {"type": "source"})

    @property
    def target_lang_url(self):
        return self.format_url(self.api_url, f"/{API_V}/languages", {"type": "target"})

    @property
    def translate_url(self):
        return self.format_url(self.api_url, f"/{API_V}/translate")

    @property
    def usage_url(self):
        return self.format_url(self.api_url, f"/{API_V}/usage")

    @property
    def headers(self):
        return {"Authorization": f"DeepL-Auth-Key {self.api_key}"}

    def init_trans(self, on_done, on_fail):
        def check_finished():
            self._init_count -= 1

            if self._init_count == 0:
                if self._init_error:
                    on_fail(self._init_error)
                else:
                    on_done()

        def on_failed(error):
            self._init_error = error
            check_finished()

        def on_languages_response(data, type_):
            try:
                trans_src = type_ == "src"
                trans_dest = type_ == "dest"

                for lang in data:
                    self.add_lang(lang["language"], lang["name"], trans_src=trans_src, trans_dest=trans_dest)

                check_finished()

            except Exception as exc:
                print(type_)
                logging.warning(exc)
                on_failed(ProviderError(ProviderErrorCode.UNEXPECTED, str(exc)))

        # Keep state of multiple request
        self._init_count = 2
        self._init_error = None

        # Request messages
        src_langs_message = self.create_message("GET", self.source_lang_url, headers=self.headers)
        dest_langs_message = self.create_message("GET", self.target_lang_url, headers=self.headers)

        # Do async requests
        self.send_and_read_and_process_response(src_langs_message, lambda d: on_languages_response(d, "src"), on_failed)
        self.send_and_read_and_process_response(
            dest_langs_message, lambda d: on_languages_response(d, "dest"), on_failed
        )

    def validate_api_key(self, key, on_done, on_fail):
        def on_response(_data):
            on_done(True)

        api_url = self.__get_api_url(key)
        url = self.format_url(api_url, f"/{API_V}/languages", {"type": "source"})
        # Headers
        headers = {"Authorization": f"DeepL-Auth-Key {key}"}
        # Request messages
        languages_message = self.create_message("GET", url, headers=headers)
        # Do async requests
        self.send_and_read_and_process_response(languages_message, on_response, on_fail)

    def translate(self, text, src, dest, on_done, on_fail):
        def on_response(data):
            try:
                translations = data.get("translations")
                detected = translations[0].get("detected_source_language")
                translation = Translation(translations[0]["text"], (text, src, dest), detected)
                on_done(translation)

            except Exception as exc:
                logging.warning(exc)
                on_fail(ProviderError(ProviderErrorCode.TRANSLATION_FAILED, str(exc)))

        # Request body
        data = {
            "text": [text],
            "target_lang": dest,
        }
        if src != "auto":
            data["source_lang"] = src

        # Request message
        message = self.create_message("POST", self.translate_url, data, self.headers)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def api_char_usage(self, on_done, on_fail):
        def on_response(data):
            try:
                usage = data.get("character_count")
                limit = data.get("character_limit")
                on_done(usage, limit)

            except Exception as exc:
                logging.warning(exc)
                on_fail(ProviderError(ProviderErrorCode.UNEXPECTED, str(exc)))

        # Request message
        message = self.create_message("GET", self.usage_url, headers=self.headers)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def cmp_langs(self, a, b):
        # Early return if both langs are just the same
        if a == b:
            return True

        # Split lang code to separate it from possible country/script code
        a_codes = a.split("-")
        b_codes = b.split("-")

        if a_codes[0] == b_codes[0]:  # Check base codes
            return True

        return False

    def check_known_errors(self, status, data):
        message = data.get("message", "") if isinstance(data, dict) else ""

        match status:
            case 403:
                if not self.api_key:
                    return ProviderError(ProviderErrorCode.API_KEY_REQUIRED, message)
                return ProviderError(ProviderErrorCode.API_KEY_INVALID, message)
            case 456:
                return ProviderError(ProviderErrorCode.SERVICE_LIMIT_REACHED, message)

        if status != 200:
            return ProviderError(ProviderErrorCode.UNEXPECTED, message)



# Copyright 2023 Mufeed Ali
# Copyright 2023 Rafael Mardojai CM
# Copyright 2023 Libretto
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gdk, GObject, Gtk

from dialect.settings import Settings


class TextView(Gtk.TextView):
    __gtype_name__ = "TextView"

    activate_mod: bool = GObject.Property(type=bool, default=True)  # type: ignore
    """If activation requieres the mod key"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set word/char text wrapping
        self.props.wrap_mode = Gtk.WrapMode.WORD_CHAR

        # Key press controller
        key_ctrlr = Gtk.EventControllerKey()
        key_ctrlr.connect("key-pressed", self._on_key_pressed)
        self.add_controller(key_ctrlr)

        # Scroll controller
        scroll_ctrlr = Gtk.EventControllerScroll(flags=Gtk.EventControllerScrollFlags.VERTICAL)
        scroll_ctrlr.connect("scroll", self._on_scroll)
        self.add_controller(scroll_ctrlr)

        # Custom font
        self._font_size = Settings.get().system_font_size
        self._font_css_provider = Gtk.CssProvider()

        # Add font CSS provider
        self.get_style_context().add_provider(self._font_css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)

    @GObject.Signal()
    def activate(self): ...

    @GObject.Property(type=int)
    def font_size(self) -> int:  # type: ignore
        return self._font_size

    @font_size.setter
    def font_size(self, value: int):
        # Save value
        self._font_size = value
        # Update CSS
        self._font_css_provider.load_from_data(f"textview {{ font-size: { str(value) }pt; }}")

    def font_size_inc(self):
        self.font_size += 5

    def font_size_dec(self):
        new_size = self.font_size - 5
        if new_size >= 6:
            self.font_size = new_size

    def _on_key_pressed(self, _ctrl, keyval: int, _keycode: int, state: Gdk.ModifierType):
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        control_mask = Gdk.ModifierType.CONTROL_MASK
        enter_keys = (Gdk.KEY_Return, Gdk.KEY_KP_Enter)

        # Activate with mod key pressed
        if control_mask == modifiers:
            if keyval in enter_keys:
                if self.activate_mod:
                    self.emit("activate")
                    return Gdk.EVENT_STOP

        # Activate without mod key pressed
        elif keyval in enter_keys:
            if not self.activate_mod:
                self.emit("activate")
                return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_scroll(self, ctrl: Gtk.EventControllerScroll, _dx: float, dy: float):
        state = ctrl.get_current_event_state()

        # If Control modifier is pressed
        if state == Gdk.ModifierType.CONTROL_MASK:
            if dy > 0:
                self.font_size_dec()
            else:
                self.font_size_inc()

            # Stop propagation
            return Gdk.EVENT_STOP

        # Propagate event (scrolled window, etc)
        return Gdk.EVENT_PROPAGATE



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from dialect.providers.base import (
    ProviderCapability,
    ProviderError,
    ProviderErrorCode,
    ProviderFeature,
    Translation,
)
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "libretranslate"
    prettyname = "LibreTranslate"

    capabilities = ProviderCapability.TRANSLATION
    features = ProviderFeature.INSTANCES | ProviderFeature.DETECTION

    defaults = {
        "instance_url": "lt.dialectapp.org",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 0

    def validate_instance(self, url, on_done, on_fail):
        def on_response(data):
            valid = False

            try:
                valid = data["info"]["title"] == "LibreTranslate"
            except:  # noqa
                pass

            on_done(valid)

        # Message request to LT API spec endpoint
        message = self.create_message("GET", self.format_url(url, "/spec"))
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail, False)

    @property
    def frontend_settings_url(self):
        return self.format_url(self.instance_url, "/frontend/settings")

    @property
    def detect_url(self):
        return self.format_url(self.instance_url, "/detect")

    @property
    def lang_url(self):
        return self.format_url(self.instance_url, "/languages")

    @property
    def suggest_url(self):
        return self.format_url(self.instance_url, "/suggest")

    @property
    def translate_url(self):
        return self.format_url(self.instance_url, "/translate")

    def init_trans(self, on_done, on_fail):
        def check_finished():
            self._init_count -= 1

            if self._init_count == 0:
                if self._init_error:
                    on_fail(self._init_error)
                else:
                    on_done()

        def on_failed(error: ProviderError):
            self._init_error = error
            check_finished()

        def on_languages_response(data):
            try:
                for lang in data:
                    self.add_lang(lang["code"], lang["name"])

                check_finished()

            except Exception as exc:
                logging.warning(exc)
                on_failed(ProviderError(ProviderErrorCode.UNEXPECTED, str(exc)))

        def on_settings_response(data):
            try:
                if data.get("suggestions", False):
                    self.features ^= ProviderFeature.SUGGESTIONS
                if data.get("apiKeys", False):
                    self.features ^= ProviderFeature.API_KEY
                if data.get("keyRequired", False):
                    self.features ^= ProviderFeature.API_KEY_REQUIRED

                self.chars_limit = int(data.get("charLimit", 0))

                check_finished()

            except Exception as exc:
                logging.warning(exc)
                on_failed(ProviderError(ProviderErrorCode.UNEXPECTED, str(exc)))

        # Keep state of multiple request
        self._init_count = 2
        self._init_error: ProviderError | None = None

        # Request messages
        languages_message = self.create_message("GET", self.lang_url)
        settings_message = self.create_message("GET", self.frontend_settings_url)

        # Do async requests
        self.send_and_read_and_process_response(languages_message, on_languages_response, on_failed)
        self.send_and_read_and_process_response(settings_message, on_settings_response, on_failed)

    def validate_api_key(self, key, on_done, on_fail):
        def on_response(data):
            valid = False
            try:
                valid = "confidence" in data[0]
            except:  # noqa
                pass

            on_done(valid)

        # Form data
        data = {
            "q": "hello",
            "api_key": key,
        }

        # Request message
        message = self.create_message("POST", self.detect_url, data, form=True)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def translate(self, text, src, dest, on_done, on_fail):
        def on_response(data):
            detected = data.get("detectedLanguage", {}).get("language", None)
            translation = Translation(data["translatedText"], (text, src, dest), detected)
            on_done(translation)

        # Request body
        data = {
            "q": text,
            "source": src,
            "target": dest,
        }
        if self.api_key and ProviderFeature.API_KEY in self.features:
            data["api_key"] = self.api_key

        # Request message
        message = self.create_message("POST", self.translate_url, data)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def suggest(self, text, src, dest, suggestion, on_done, on_fail):
        def on_response(data):
            on_done(data.get("success", False))

        # Form data
        data = {
            "q": text,
            "source": src,
            "target": dest,
            "s": suggestion,
        }
        if self.api_key and ProviderFeature.API_KEY in self.features:
            data["api_key"] = self.api_key

        # Request message
        message = self.create_message("POST", self.suggest_url, data, form=True)
        # Do async request
        self.send_and_read_and_process_response(message, on_response, on_fail)

    def check_known_errors(self, _status, data):
        if not data:
            return ProviderError(ProviderErrorCode.EMPTY, "Response is empty!")
        if "error" in data:
            error = data["error"]

            if error == "Please contact the server operator to obtain an API key":
                return ProviderError(ProviderErrorCode.API_KEY_REQUIRED, error)
            elif error == "Invalid API key":
                return ProviderError(ProviderErrorCode.API_KEY_INVALID, error)
            elif "is not supported" in error:
                return ProviderError(ProviderErrorCode.INVALID_LANG_CODE, error)
            elif "exceeds text limit" in error:
                return ProviderError(ProviderErrorCode.BATCH_SIZE_EXCEEDED, error)
            elif "exceeds character limit" in error:
                return ProviderError(ProviderErrorCode.CHARACTERS_LIMIT_EXCEEDED, error)
            elif "Cannot translate text" in error or "format is not supported" in error:
                return ProviderError(ProviderErrorCode.TRANSLATION_FAILED, error)
            else:
                return ProviderError(ProviderErrorCode.UNEXPECTED, error)



# Copyright 2024 Mufeed Ali
# Copyright 2024 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from typing import TypedDict

from gi.repository import Gio, GLib, Secret

from dialect.define import APP_ID

SECRETS_SCHEMA = Secret.Schema.new(
    APP_ID,
    Secret.SchemaFlags.NONE,
    {
        "provider": Secret.SchemaAttributeType.STRING,
    },
)


class ProviderDefaults(TypedDict):
    instance_url: str
    api_key: str
    src_langs: list[str]
    dest_langs: list[str]


class ProviderSettings(Gio.Settings):
    """
    Helper class for providers settings
    """

    def __init__(self, name: str, defaults: ProviderDefaults):
        super().__init__(schema_id=f"{APP_ID}.translator", path=f"/app/drey/Dialect/translators/{name}/")

        self.name = name
        self.defaults = defaults  # set of per-provider defaults
        self._secret_attr = {"provider": name}

    @property
    def instance_url(self) -> str:
        """Instance url."""
        return self.get_string("instance-url") or self.defaults["instance_url"]

    @instance_url.setter
    def instance_url(self, url: str):
        self.set_string("instance-url", url)

    @property
    def api_key(self) -> str:
        """API key."""

        # Check if we have an old API KEY in GSettings for migration
        if gsettings := self.get_string("api-key"):
            self.api_key = gsettings  # Save to secret
            self.set_string("api-key", "")  # Clear GSetting
            return gsettings

        try:
            return Secret.password_lookup_sync(SECRETS_SCHEMA, self._secret_attr, None) or self.defaults["api_key"]
        except GLib.Error as exc:
            logging.warning(exc)

        return self.defaults["api_key"]

    @api_key.setter
    def api_key(self, api_key: str):
        try:
            if api_key:
                Secret.password_store_sync(
                    SECRETS_SCHEMA,
                    self._secret_attr,
                    Secret.COLLECTION_DEFAULT,
                    f"Dialect provider API KEY for {self.name}",
                    api_key,
                    None,
                )
            else:  # Remove secret
                Secret.password_clear_sync(SECRETS_SCHEMA, self._secret_attr, None)

            # Fake change in api-key setting
            self.emit("changed::api-key", "api-key")

        except GLib.Error as exc:
            logging.warning(exc)

    @property
    def src_langs(self) -> list[str]:
        """Recent source languages."""
        return self.get_strv("src-langs") or self.defaults["src_langs"]

    @src_langs.setter
    def src_langs(self, src_langs: list[str]):
        self.set_strv("src-langs", src_langs)

    @property
    def dest_langs(self) -> list[str]:
        """Recent destination languages."""
        return self.get_strv("dest-langs") or self.defaults["dest_langs"]

    @dest_langs.setter
    def dest_langs(self, dest_langs: list[str]):
        self.set_strv("dest-langs", dest_langs)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# Copyright 2023 Libretto
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Literal

from gi.repository import Gio, GLib, GObject

from dialect.define import APP_ID
from dialect.providers import (
    TTS,
    check_translator_availability,
    get_fallback_translator_name,
)


class Settings(Gio.Settings):
    """
    Dialect settings handler
    """

    instance = None

    def __init__(self, *args):
        super().__init__(*args)

        self._translators = self.get_child("translators")
        self._tts = self.get_child("tts")

    @staticmethod
    def new() -> Settings:
        """Create a new instance of Settings."""
        g_settings = Settings(APP_ID)
        return g_settings

    @staticmethod
    def get() -> Settings:
        """Return an active instance of Settings."""
        if Settings.instance is None:
            Settings.instance = Settings.new()
        return Settings.instance

    @GObject.Signal(arg_types=(str,))
    def translator_changed(self, _name: str): ...

    @GObject.Signal(arg_types=(str,))
    def tts_changed(self, _name: str): ...

    @property
    def translators_list(self) -> list[str]:
        return self._translators.get_strv("list")

    @translators_list.setter
    def translators_list(self, translators: list[str]):
        self._translators.set_strv("list", translators)

    @property
    def active_translator(self) -> str:
        value = self._translators.get_string("active")

        if check_translator_availability(value):
            return value

        self.active_translator = get_fallback_translator_name()
        return get_fallback_translator_name()

    @active_translator.setter
    def active_translator(self, translator: str):
        self._translators.set_string("active", translator)
        self.emit("translator-changed", translator)

    @property
    def window_size(self) -> tuple[int, int]:
        value = self.get_value("window-size")
        return (value[0], value[1])

    @window_size.setter
    def window_size(self, size: tuple[int, int]):
        width, height = size
        self.set_value("window-size", GLib.Variant("ai", [width, height]))

    @property
    def translate_accel(self) -> str:
        """Return the user's preferred translation shortcut."""
        value = self.translate_accel_value

        if value == 0:
            return "<Primary>Return"
        if value == 1:
            return "Return"

        return "<Primary>Return"

    @property
    def translate_accel_value(self) -> int:
        """Return the user's preferred translation shortcut value."""
        return self.get_int("translate-accel")

    @property
    def custom_default_font_size(self) -> bool:
        """Return whether the user wants a custom default font size."""
        return self.get_boolean("custom-default-font-size")

    @custom_default_font_size.setter
    def custom_default_font_size(self, enabled: bool):
        self.set_boolean("custom-default-font-size", enabled)

    @property
    def default_font_size(self) -> int:
        """Return the user's preferred default font size."""
        return self.get_int("default-font-size")

    @default_font_size.setter
    def default_font_size(self, size: int):
        self.set_int("default-font-size", size)

    @property
    def system_font_size(self) -> int:
        """Return the systems's default font size."""
        from gi.repository import Gtk

        try:
            gtk_font_name = Gtk.Settings.get_default().get_property("gtk-font-name").split(" ")  # type: ignore
            return int(gtk_font_name[-1])
        except Exception:
            return 11

    @property
    def active_tts(self) -> str:
        """Return the user's preferred TTS service."""
        value = self._tts.get_string("active")

        if value != "" and value not in TTS.keys():
            value = ""
            self.active_tts = value

        return value

    @active_tts.setter
    def active_tts(self, tts: str):
        """Set the user's preferred TTS service."""
        self._tts.set_string("active", tts)
        self.emit("tts-changed", tts)

    @property
    def color_scheme(self) -> str:
        return self.get_string("color-scheme")

    @color_scheme.setter
    def color_scheme(self, scheme: Literal["auto", "light", "dark"]):
        self.set_string("color-scheme", scheme)

    @property
    def live_translation(self) -> bool:
        return self.get_boolean("live-translation")

    @live_translation.setter
    def live_translation(self, state: bool):
        self.set_boolean("live-translation", state)

    @property
    def sp_translation(self) -> bool:
        return self.get_boolean("sp-translation")

    @sp_translation.setter
    def sp_translation(self, state: bool):
        self.set_boolean("sp-translation", state)

    @property
    def show_pronunciation(self) -> bool:
        return self.get_boolean("show-pronunciation")

    @property
    def show_pronunciation_value(self) -> GLib.Variant:
        return self.get_value("show-pronunciation")

    @show_pronunciation.setter
    def show_pronunciation(self, state: bool):
        self.set_boolean("show-pronunciation", state)

    @property
    def src_auto(self) -> bool:
        return self.get_boolean("src-auto")

    @src_auto.setter
    def src_auto(self, state: bool):
        self.set_boolean("src-auto", state)



# Copyright 2022 Mufeed Ali
# Copyright 2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging
from typing import Callable

from gi.repository import Gio, GLib, Soup

from dialect.providers.base import BaseProvider, ProviderError, ProviderErrorCode
from dialect.session import Session


class SoupProvider(BaseProvider):
    """Base class for providers needing libsoup helpers"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def encode_data(self, data) -> GLib.Bytes | None:
        """
        Convert Python data to JSON and bytes.

        Args:
            data: Data to encode, anything json.dumps can handle
        """
        data_glib_bytes = None
        try:
            data_bytes = json.dumps(data).encode("utf-8")
            data_glib_bytes = GLib.Bytes.new(data_bytes)
        except Exception as exc:
            logging.warning(exc)
        return data_glib_bytes

    def create_message(self, method: str, url: str, data={}, headers: dict = {}, form: bool = False) -> Soup.Message:
        """
        Create a libsoup's message.

        Encodes data and adds it to the message as the request body.
        If form is true, data is encoded as application/x-www-form-urlencoded.

        Args:
            method: HTTP method of the message
            url: Url of the message
            data: Request body or form data
            headers: HTTP headers of the message
            form: If the data should be encoded as a form
        """

        if form and data:
            form_data = Soup.form_encode_hash(data)
            message = Soup.Message.new_from_encoded_form(method, url, form_data)
        else:
            message = Soup.Message.new(method, url)

        if message:
            if data and not form:
                data = self.encode_data(data)
                message.set_request_body_from_bytes("application/json", data)
            if headers:
                for name, value in headers.items():
                    message.get_request_headers().append(name, value)
            if "User-Agent" not in headers:
                message.get_request_headers().append("User-Agent", "Dialect App")

        return message  # type: ignore

    def send_and_read(self, message: Soup.Message, callback: Callable[[Session, Gio.AsyncResult], None]):
        """
        Helper method for libsoup's send_and_read_async.

        Useful when priority and cancellable is not needed.

        Args:
            message: Message to send
            callback: Callback called from send_and_read_async to finish request
        """
        Session.get().send_and_read_async(message, 0, None, callback)

    def read_data(self, data: bytes | None) -> dict:
        """
        Get JSON data from bytes.

        Args:
            data: Bytes to read
        """
        return json.loads(data) if data else {}

    def read_response(self, session: Session, result: Gio.AsyncResult) -> dict:
        """
        Get JSON data from session result.

        Finishes request from send_and_read_async and gets body dict.

        Args:
            session: Session where the request wa sent
            result: Result of send_and_read_async callback
        """
        response = session.get_response(session, result)
        return self.read_data(response)

    def check_known_errors(self, status: Soup.Status, data: dict | bytes | None) -> None | ProviderError:
        """
        Checks data for possible response errors and return a found error if any.

        This should be implemented by subclases.

        Args:
            data: Response body data
        """
        return None

    def process_response(
        self,
        session: Session,
        result: Gio.AsyncResult,
        message: Soup.Message,
        on_continue: Callable[[dict | bytes | None], None],
        on_fail: Callable[[ProviderError], None],
        check_common: bool = True,
        json: bool = True,
    ):
        """
        Helper method for the most common workflow for processing soup responses.

        Checks for soup errors, then checks for common errors on data and calls on_fail
        if any, otherwise calls on_continue where the provider will finish the process.

        If json is false check_common is ignored and the data isn't processed as JSON and bites are passed to
        on_continue.

        Args:
            session: Session where the request wa sent
            result: Result of send_and_read_async callback
            message: The message that was sent
            on_continue: Called after data was got successfully
            on_fail: Called after any error on request or in check_known_errors
            check_common: If response data should be checked for errors using check_known_errors
            json: If data should be processed as JSON using read_response
        """

        try:
            if json:
                data = self.read_response(session, result)
            else:
                data = Session.get_response(session, result)

            if check_common:
                error = self.check_known_errors(message.get_status(), data)
                if error:
                    on_fail(error)
                    return

            on_continue(data)

        except Exception as exc:
            logging.warning(exc)
            on_fail(ProviderError(ProviderErrorCode.NETWORK, str(exc)))

    def send_and_read_and_process_response(
        self,
        message: Soup.Message,
        on_continue: Callable[[dict | bytes | None], None],
        on_fail: Callable[[ProviderError], None],
        check_common: bool = True,
        json: bool = True,
    ):
        """
        Helper packaging send_and_read and process_response.

        Avoids providers having to deal with many callbacks.

        message: Message to send
        on_continue: Called after data was got successfully
        on_fail: Called after any error on request or in check_known_errors
        check_common: If response data should be checked for errors using check_known_errors
        json: If data should be processed as JSON using read_response
        """

        def on_response(session: Session, result: Gio.AsyncResult):
            self.process_response(session, result, message, on_continue, on_fail, check_common, json)

        self.send_and_read(message, on_response)
